from typing import Callable
import logging
import torch.utils.data
import numpy as np
import pytorch_lightning as pl
from itertools import islice

import webdataset as wds

from galaxy_datasets import transforms

# https://github.com/webdataset/webdataset-lightning/blob/main/train.py
class WebDataModule(pl.LightningDataModule):
    def __init__(
            self,
            train_urls=None,
            val_urls=None,
            test_urls=None,
            predict_urls=None,
            label_cols=None,
            # hardware
            batch_size=64,
            num_workers=4,
            prefetch_factor=4,
            cache_dir=None,
            train_transform: Callable=None,
            inference_transform: Callable=None
            ):
        super().__init__()

        self.train_urls = train_urls
        self.val_urls = val_urls
        self.test_urls = test_urls
        self.predict_urls = predict_urls

        self.set_dataset_size_attributes(train_urls, val_urls, test_urls, predict_urls)

        self.label_cols = label_cols

        self.batch_size = batch_size
        self.num_workers = num_workers
        self.prefetch_factor = prefetch_factor

        self.cache_dir = cache_dir

        self.train_transform = train_transform
        self.inference_transform = inference_transform

        for url_name in ['train', 'val', 'test', 'predict']:
            urls = getattr(self, f'{url_name}_urls')
            if urls is not None:
                logging.info(f"{url_name} (before hardware splits) = {len(urls)} e.g. {urls[0]}", )

        logging.info(f"batch_size: {self.batch_size}, num_workers: {self.num_workers}")

    def set_dataset_size_attributes(self, train_urls=None, val_urls=None, test_urls=None, predict_urls=None):
        if train_urls is not None:
            # assume the size of each shard is encoded in the filename as ..._{size}.tar
            self.train_size = interpret_dataset_size_from_urls(train_urls)
        if val_urls is not None:
            self.val_size = interpret_dataset_size_from_urls(val_urls)
        if test_urls is not None:
            self.test_size = interpret_dataset_size_from_urls(test_urls)
        if predict_urls is not None:
            self.predict_size = interpret_dataset_size_from_urls(predict_urls)

    # def make_image_transform(self, mode="train"):
    #     # only used if you don't explicitly pass a transform to the datamodule
    # NOW DEPRECATED because we don't use albumentations

    #     augmentation_transform = transforms.default_transforms(
    #         crop_scale_bounds=self.crop_scale_bounds,
    #         crop_ratio_bounds=self.crop_ratio_bounds,
    #         resize_after_crop=self.resize_after_crop,
    #         pytorch_greyscale=self.greyscale,
    #         to_float=False,  # True was wrong, webdataset rgb decoder already converts to 0-1 float
    #         # TODO now changed on dev branch will be different for new model training runs
    #         # this compose will now return a Tensor object, default transform was updated
    #         to_tensor=True
    #     )  # A.Compose object

    #     # logging.warning('Minimal augmentations for speed test')
    #     # augmentation_transform = transforms.fast_transforms(
    #     #     resize_after_crop=self.resize_after_crop,
    #     #     pytorch_greyscale=not self.color
    #     # )  # A.Compose object

    #     def do_transform(img):
    #         # img is 0-1 np array, intended for albumentations
    #         assert img.shape[2] < 4  # 1 or 3 channels in shape[2] dim, i.e. numpy/pil HWC convention
    #         # if not, check decode mode is 'rgb' not 'torchrgb'
    #         # default augmentation now returns CHW tensor
    #         return augmentation_transform(image=np.array(img))["image"]
    #     return do_transform


    def make_loader(self, urls, mode="train"):
        logging.info('Making loader with mode {}'.format(mode))

        dataset_size = getattr(self, f'{mode}_size')
        if mode == "train":
            shuffle = min(dataset_size, 5000)
        else:
            assert mode in ['val', 'test', 'predict'], mode
            shuffle = 0

        if self.train_transform is None:
            logging.info('Using default transform')
            raise NotImplementedError('Deprecated')
            # transform_image = self.make_image_transform()
        # else:
            # logging.info('Ignoring other arguments to WebDataModule and using directly-passed transforms')

        transform_image = self.train_transform if mode == 'train' else self.inference_transform

        transform_label = dict_to_label_cols_factory(self.label_cols)

        dataset =  wds.WebDataset(urls, cache_dir=self.cache_dir, shardshuffle=shuffle>0, nodesplitter=nodesplitter_func)
        # https://webdataset.github.io/webdataset/multinode/ 
        # WDS 'knows' which worker it is running on and selects a subset of urls accordingly
           
        if shuffle > 0:
            dataset = dataset.shuffle(shuffle)

        # this controls how webdataset decodes the images, either rgb or torch tensors, see above
        decode_mode = 'torchrgb'  # tensor, for torchvision. Set pil_to_tensor=False in torchvision transforms, already tensor
        # decode_mode = 'rgb' # loads 0-1 np.array, for albumentations
        dataset = dataset.decode(decode_mode)

        # now the webdataset needs to be unpacked (into tensor image, or tensor image, tensor label)
        # and transformed with whatever you passed to transform_image
    
        if mode == 'predict':
            if self.label_cols != ['id_str']:
                logging.info('Will return images only')
                # dataset = dataset.extract_keys('image.jpg').map(transform_image)
                dataset = dataset.to_tuple('image.jpg').map_tuple(transform_image)  # (im,) tuple. But map applied to all elements
                # .map(get_first)
            else:
                logging.info('Will return id_str only')
                dataset = dataset.to_tuple('__key__')
        else:
            
            dataset = (
                dataset.to_tuple('image.jpg', 'labels.json')
                .map_tuple(transform_image, transform_label)
            )

        # torch collate stacks dicts nicely while webdataset only lists them
        # so use the torch collate instead
        dataset = dataset.batched(self.batch_size, torch.utils.data.default_collate, partial=False) 

        # temp hack instead
        if mode in ['train', 'val']:
            assert dataset_size % self.batch_size == 0, (dataset_size, self.batch_size, dataset_size % self.batch_size)
        # for test/predict, always single GPU anyway

        # if mode == "train":
            # ensure same number of batches in all clients
            # loader = loader.ddp_equalize(dataset_size // self.batch_size)
            # print("# loader length", len(loader))

        loader = webdataset_to_webloader(dataset, self.num_workers, self.prefetch_factor)

        return loader

    def train_dataloader(self):
        return self.make_loader(self.train_urls, mode="train")

    def val_dataloader(self):
        return self.make_loader(self.val_urls, mode="val")

    def test_dataloader(self):
        return self.make_loader(self.test_urls, mode="test")
    
    def predict_dataloader(self):
        return self.make_loader(self.predict_urls, mode="predict")

def identity(x):
    return x

def nodesplitter_func(urls):
    urls_to_use = list(wds.split_by_node(urls))  # rely on WDS for the hard work
    rank, world_size, worker, num_workers = wds.utils.pytorch_worker_info()
    logging.debug(
        f'''
        Splitting urls within webdatamodule with WORLD_SIZE: 
        {world_size}, RANK: {rank}, WORKER: {worker} of {num_workers}\n
        URLS: {len(urls_to_use)} (e.g. {urls_to_use[0]})\n\n)
        '''
        )
    return urls_to_use

def interpret_shard_size_from_url(url):
    assert isinstance(url, str), TypeError(url)
    return int(url.rstrip('.tar').split('_')[-1])

def interpret_dataset_size_from_urls(urls):
    return sum([interpret_shard_size_from_url(url) for url in urls])

def get_first(x):
    return x[0]

def custom_collate(x):
    if isinstance(x, list) and len(x) == 1:
        x = x[0]
    return torch.utils.data.default_collate(x)


def webdataset_to_webloader(dataset, num_workers, prefetch_factor):
    loader = wds.WebLoader(
            dataset,
            batch_size=None,  # already batched
            shuffle=False,  # already shuffled
            num_workers=num_workers,
            pin_memory=True,
            prefetch_factor=prefetch_factor
        )

    # loader.length = dataset_size // batch_size
    return loader


def dict_to_label_cols_factory(label_cols=None):
    # converts from dict to vector of counts
    if label_cols is not None:
        def label_transform(label_dict):
            return torch.from_numpy(np.array([label_dict.get(col, 0) for col in label_cols])).double()  # gets cast to int in zoobot loss
        return label_transform
    else:
        return identity  # do nothing


# Used for hybrid pretraining
def dict_to_filled_dict_factory(label_cols):
    logging.info(f'label cols: {label_cols}')
    # might be a little slow, but very safe
    def label_transform(label_dict: dict):

        # modifies inplace with 0 iff key missing
        # [label_dict.setdefault(col, 0) for col in label_cols]

        for col in label_cols:
            label_dict[col] = label_dict.get(col, 0)

        # label_dict_with_default = defaultdict(0)
        # label_dict_with_default.update(label_dict)

        return label_dict
    return label_transform


# just for debugging
def load_wds_with_webdatamodule(save_loc, label_cols, batch_size=16, max_to_load=3):
    wdm = WebDataModule(
        train_urls=save_loc,
        val_urls=save_loc,  # not used
        # train_size=len(train_catalog),
        # val_size=0,
        label_cols=label_cols,
        num_workers=1,
        batch_size=batch_size
    )
    wdm.setup('fit')

    if max_to_load is not None:
        sample_iterator =islice(wdm.train_dataloader(), 0, max_to_load)
    else:
        sample_iterator = wdm.train_dataloader()
    for sample in sample_iterator:
        images, labels = sample
        logging.info(images.shape)
        # logging.info(len(labels))  # list of dicts
        logging.info(labels.shape)




import os
import logging
from typing import Tuple, List

import pandas as pd

# TODO could eventually refactor this out of Zoobot as well
from galaxy_datasets.shared import label_metadata

from galaxy_datasets.shared import download_utils
from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST

if not INTERNAL_URLS_EXIST:
    raise FileNotFoundError
from galaxy_datasets.shared import internal_urls

"""
This downloads the *labelled* GZ DESI galaxies, at all redshifts, with no division between data releases (except label_cols)
useful to conveniently download galaxies for GZ DESI itself (see gz-decals-classifiers, though this also uses catalogs directly)
or for other large-scale supervised learning problems or for pretraining ahead of active learning
does not include unlabelled galaxies (for e.g. contrastive learning, active learning)

Added to make GZ DESI easily downloadable, basically
"""
def gz_desi(root, train, download) -> Tuple[pd.DataFrame, List]:
    logging.info('Setting up gz_desi dataset')
    resources = [
        (internal_urls.gz_desi_train_catalog, 'edcbbe4f0adb767d26bdc5de26aa854d'),
        (internal_urls.gz_desi_test_catalog, '95bfed06c5a00742f58171f46cb1d8c1'),
        (internal_urls.gz_desi_images_chunk_00, '76c861762633d022603686e0794ce99c'),
        (internal_urls.gz_desi_images_chunk_01, '844502e49d6e087b6c1f4fcfef0772c0'),
        (internal_urls.gz_desi_images_chunk_02, 'b793771b8fc9383b616d2dc67bdfbf7e'),
        (internal_urls.gz_desi_images_chunk_03, '003821161b4990a49d1dcb5d72d49b12'),
        (internal_urls.gz_desi_images_chunk_04, 'aa85dec158d95edea98b31efe30231d5'),
        (internal_urls.gz_desi_images_chunk_05, '36ed5893441a533d1d74fe480cb42cb2'),
        (internal_urls.gz_desi_images_chunk_06, '6b84e92c7b4f1edbb372d2247ff1740a'),
        (internal_urls.gz_desi_images_chunk_07, 'feef448b45eeb3d5edc3f44d936f386a'),
        (internal_urls.gz_desi_images_chunk_08, 'c8be005044f94046a7080617815a3d00')
    ]
    images_to_spotcheck = ['331000/331000_345.jpg']  # newly added

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck, archive_includes_subdir=False)
    if download is True:
        downloader.download()


    # useful_columns = label_cols + ['subfolder', 'filename']
    if train:
        train_catalog_loc = os.path.join(root, 'gz_desi_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc)
    else:
        test_catalog_loc = os.path.join(root, 'gz_desi_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc)

    # removed 'root' from here as downloader.image_dir already includes root
    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    # default, but not actually used here
    label_cols = label_metadata.decals_all_campaigns_ortho_label_cols
    logging.info('gz_desi dataset ready')
    return catalog, label_cols




import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils


def galaxy_mnist(root, train, download):
    logging.info('Setting up galaxy_mnist dataset')
    resources = [
        ('https://dl.dropboxusercontent.com/s/6xlym0ney5q9aec14vvy1/galaxy_mnist_train_catalog.parquet?rlkey=fc5knqscwk2h6r4z5d3156vzr&dl=0', 'cd22b21d165802f4bc1adf997424aec2'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/9assxy247i1nq8wjy7o0a/galaxy_mnist_test_catalog.parquet?rlkey=y1ns24xw0tcyntodi61rv14fu&dl=0', 'c5cca8d8afb6fb0d59baeb310a35d594'),  # test catalog
        ('https://dl.dropboxusercontent.com/s/lj89307kjx5plme9f66js/galaxy_mnist_images.tar.gz?rlkey=gvlwx2rl3hqpo3gplb9zlqrz0', '1c6cb0447f2f7ed676c3363ee194ced9')  # the images
    ]
    images_to_spotcheck = [
        'images/92abc34c-824c-4d46-a79c-cc28a768149f.jpg',
        'images/a4bd154c-327c-4064-90d2-7d3d2bd045dc.jpg'
    ]

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    if train:
        train_catalog_loc = os.path.join(root, 'galaxy_mnist_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc)
    else:
        test_catalog_loc = os.path.join(root, 'galaxy_mnist_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['filename']), axis=1)

    label_cols = ['label']  # 0-3, see https://github.com/mwalmsley/galaxy_mnist

    logging.info('galaxy_mnist dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    df, _ = galaxy_mnist(root='/home/walml/repos/galaxy-datasets/roots/galaxy_mnist', train=True, download=True)
    example_file = df['file_loc'][0]
    assert os.path.isfile(example_file), example_file



import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils


def demo_gz_candels(root, train, download):
    logging.info('Setting up demo_gz_candels dataset')
    resources = [
        ('https://dl.dropboxusercontent.com/s/rdejdbzn3vsid3r/demo_gz_candels_train_catalog.parquet', '586632f9f832748dec2a52a5375bf49e'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/2dn018co3ip7f8t/demo_gz_candels_test_catalog.parquet',  '2519671bfac59a84d971402d0ab8094c'), # test catalog
        ('https://dl.dropboxusercontent.com/s/1nhhvxsi53rjx7a/demo_gz_candels_images.tar.gz', 'fc2d856cc8d780b381b44681cadf57d6')  # the images
    ]
    images_to_spotcheck = [
        'images/COS_27104.jpg',
        'images/UDS_13502.jpg'
    ]

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    if train:
        train_catalog_loc = os.path.join(root, 'demo_gz_candels_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc)
    else:
        test_catalog_loc = os.path.join(root, 'demo_gz_candels_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['filename']), axis=1)

    label_cols = ['ring']  # 1 or 0

    logging.info('demo_gz_candels dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    demo_gz_candels(root='/Users/user/repos/galaxy-datasets/roots/demo_gz_candels', train=True, download=True)


from typing import Any, Callable, Dict, IO, Iterable, Iterator, List, Optional, Tuple
import os
import gzip
import bz2
import lzma
import tarfile
import zipfile
import pathlib
import sys
import hashlib
import urllib
import requests
import contextlib

import warnings
import re
import itertools
from tqdm import tqdm
from urllib.parse import urlparse


# https://github.com/pytorch/vision/blob/main/torchvision/datasets/utils.py#L433
def download_and_extract_archive(
    url: str,
    download_root: str,
    extract_root: Optional[str] = None,
    filename: Optional[str] = None,
    md5: Optional[str] = None,
    remove_finished: bool = False,
) -> None:
    download_root = os.path.expanduser(download_root)
    if extract_root is None:
        extract_root = download_root
    if not filename:
        filename = os.path.basename(url)

    download_url(url, download_root, filename, md5)

    archive = os.path.join(download_root, filename)
    print(f"Extracting {archive} to {extract_root}")
    extract_archive(archive, extract_root, remove_finished)


# https://github.com/pytorch/vision/blob/main/torchvision/datasets/utils.py#L119
def download_url(
    url: str, root: str, filename: Optional[str] = None, md5: Optional[str] = None, max_redirect_hops: int = 3
) -> None:
    """Download a file from a url and place it in root.
    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the basename of the URL
        md5 (str, optional): MD5 checksum of the download. If None, do not check
        max_redirect_hops (int, optional): Maximum number of redirect hops allowed
    """
    root = os.path.expanduser(root)
    if not filename:
        filename = os.path.basename(url)
    fpath = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    # check if file is already present locally
    if check_integrity(fpath, md5):
        print("Using downloaded and verified file: " + fpath)
        return

    if _is_remote_location_available():
        _download_file_from_remote_location(fpath, url)
    else:
        # expand redirect chain if needed
        url = _get_redirect_url(url, max_hops=max_redirect_hops)

        # check if file is located on Google Drive
        file_id = _get_google_drive_file_id(url)
        if file_id is not None:
            return download_file_from_google_drive(file_id, root, filename, md5)

        # download the file
        try:
            print("Downloading " + url + " to " + fpath)
            _urlretrieve(url, fpath)
        except (urllib.error.URLError, OSError) as e:  # type: ignore[attr-defined]
            if url[:5] == "https":
                url = url.replace("https:", "http:")
                print("Failed download. Trying https -> http instead. Downloading " + url + " to " + fpath)
                _urlretrieve(url, fpath)
            else:
                raise e

    # check integrity of downloaded file
    if not check_integrity(fpath, md5):
        raise RuntimeError("File not found or corrupted.")



# https://github.com/pytorch/vision/blob/main/torchvision/datasets/utils.py#L397
def extract_archive(from_path: str, to_path: Optional[str] = None, remove_finished: bool = False) -> str:
    """Extract an archive.
    The archive type and a possible compression is automatically detected from the file name. If the file is compressed
    but not an archive the call is dispatched to :func:`decompress`.
    Args:
        from_path (str): Path to the file to be extracted.
        to_path (str): Path to the directory the file will be extracted to. If omitted, the directory of the file is
            used.
        remove_finished (bool): If ``True``, remove the file after the extraction.
    Returns:
        (str): Path to the directory the file was extracted to.
    """
    if to_path is None:
        to_path = os.path.dirname(from_path)

    suffix, archive_type, compression = _detect_file_type(from_path)
    if not archive_type:
        return _decompress(
            from_path,
            os.path.join(to_path, os.path.basename(from_path).replace(suffix, "")),
            remove_finished=remove_finished,
        )

    # We don't need to check for a missing key here, since this was already done in _detect_file_type()
    extractor = _ARCHIVE_EXTRACTORS[archive_type]

    extractor(from_path, to_path, compression)
    if remove_finished:
        os.remove(from_path)

    return 


def _detect_file_type(file: str) -> Tuple[str, Optional[str], Optional[str]]:
    """Detect the archive type and/or compression of a file.
    Args:
        file (str): the filename
    Returns:
        (tuple): tuple of suffix, archive type, and compression
    Raises:
        RuntimeError: if file has no suffix or suffix is not supported
    """
    suffixes = pathlib.Path(file).suffixes
    if not suffixes:
        raise RuntimeError(
            f"File '{file}' has no suffixes that could be used to detect the archive type and compression."
        )
    suffix = suffixes[-1]

    # check if the suffix is a known alias
    if suffix in _FILE_TYPE_ALIASES:
        return (suffix, *_FILE_TYPE_ALIASES[suffix])

    # check if the suffix is an archive type
    if suffix in _ARCHIVE_EXTRACTORS:
        return suffix, suffix, None

    # check if the suffix is a compression
    if suffix in _COMPRESSED_FILE_OPENERS:
        # check for suffix hierarchy
        if len(suffixes) > 1:
            suffix2 = suffixes[-2]

            # check if the suffix2 is an archive type
            if suffix2 in _ARCHIVE_EXTRACTORS:
                return suffix2 + suffix, suffix2, suffix

        return suffix, None, suffix

    valid_suffixes = sorted(set(_FILE_TYPE_ALIASES) | set(_ARCHIVE_EXTRACTORS) | set(_COMPRESSED_FILE_OPENERS))
    raise RuntimeError(f"Unknown compression or archive type: '{suffix}'.\nKnown suffixes are: '{valid_suffixes}'.")




def _decompress(from_path: str, to_path: Optional[str] = None, remove_finished: bool = False) -> str:
    r"""Decompress a file.
    The compression is automatically detected from the file name.
    Args:
        from_path (str): Path to the file to be decompressed.
        to_path (str): Path to the decompressed file. If omitted, ``from_path`` without compression extension is used.
        remove_finished (bool): If ``True``, remove the file after the extraction.
    Returns:
        (str): Path to the decompressed file.
    """
    suffix, archive_type, compression = _detect_file_type(from_path)
    if not compression:
        raise RuntimeError(f"Couldn't detect a compression from suffix {suffix}.")

    if to_path is None:
        to_path = from_path.replace(suffix, archive_type if archive_type is not None else "")

    # We don't need to check for a missing key here, since this was already done in _detect_file_type()
    compressed_file_opener = _COMPRESSED_FILE_OPENERS[compression]

    with compressed_file_opener(from_path, "rb") as rfh, open(to_path, "wb") as wfh:
        wfh.write(rfh.read())

    if remove_finished:
        os.remove(from_path)

    return to_path


def _extract_tar(from_path: str, to_path: str, compression: Optional[str]) -> None:
    with tarfile.open(from_path, f"r:{compression[1:]}" if compression else "r") as tar:
        tar.extractall(to_path)

def _extract_zip(from_path: str, to_path: str, compression: Optional[str]) -> None:
    with zipfile.ZipFile(
        from_path, "r", compression=_ZIP_COMPRESSION_MAP[compression] if compression else zipfile.ZIP_STORED
    ) as zip:
        zip.extractall(to_path)



_ARCHIVE_EXTRACTORS: Dict[str, Callable[[str, str, Optional[str]], None]] = {
    ".tar": _extract_tar,
    ".zip": _extract_zip,
}
_COMPRESSED_FILE_OPENERS: Dict[str, Callable[..., IO]] = {
    ".bz2": bz2.open,
    ".gz": gzip.open,
    ".xz": lzma.open,
}
_FILE_TYPE_ALIASES: Dict[str, Tuple[Optional[str], Optional[str]]] = {
    ".tbz": (".tar", ".bz2"),
    ".tbz2": (".tar", ".bz2"),
    ".tgz": (".tar", ".gz"),
}

_ZIP_COMPRESSION_MAP: Dict[str, int] = {
    ".bz2": zipfile.ZIP_BZIP2,
    ".xz": zipfile.ZIP_LZMA,
}
USER_AGENT = "pytorch/vision"


def check_integrity(fpath: str, md5: Optional[str] = None) -> bool:
    if not os.path.isfile(fpath):
        return False
    if md5 is None:
        return True
    return check_md5(fpath, md5)



def calculate_md5(fpath: str, chunk_size: int = 1024 * 1024) -> str:
    # Setting the `usedforsecurity` flag does not change anything about the functionality, but indicates that we are
    # not using the MD5 checksum for cryptography. This enables its usage in restricted environments like FIPS. Without
    # it torchvision.datasets is unusable in these environments since we perform a MD5 check everywhere.
    if sys.version_info >= (3, 9):
        md5 = hashlib.md5(usedforsecurity=False)
    else:
        md5 = hashlib.md5()
    with open(fpath, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            md5.update(chunk)
    return md5.hexdigest()


def check_md5(fpath: str, md5: str, **kwargs: Any) -> bool:
    return md5 == calculate_md5(fpath, **kwargs)

def _is_remote_location_available() -> bool:
    return False

def _download_file_from_remote_location(fpath: str, url: str) -> None:
    pass


def _get_redirect_url(url: str, max_hops: int = 3) -> str:
    initial_url = url
    headers = {"Method": "HEAD", "User-Agent": USER_AGENT}

    for _ in range(max_hops + 1):
        with urllib.request.urlopen(urllib.request.Request(url, headers=headers)) as response:
            if response.url == url or response.url is None:
                return url

            url = response.url
    else:
        raise RecursionError(
            f"Request to {initial_url} exceeded {max_hops} redirects. The last redirect points to {url}."
        )


def download_file_from_google_drive(file_id: str, root: str, filename: Optional[str] = None, md5: Optional[str] = None):
    """Download a Google Drive file from  and place it in root.
    Args:
        file_id (str): id of file to be downloaded
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the id of the file.
        md5 (str, optional): MD5 checksum of the download. If None, do not check
    """
    # Based on https://stackoverflow.com/questions/38511444/python-download-files-from-google-drive-using-url

    root = os.path.expanduser(root)
    if not filename:
        filename = file_id
    fpath = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    if check_integrity(fpath, md5):
        print(f"Using downloaded {'and verified ' if md5 else ''}file: {fpath}")
        return

    url = "https://drive.google.com/uc"
    params = dict(id=file_id, export="download")
    with requests.Session() as session:
        response = session.get(url, params=params, stream=True)

        for key, value in response.cookies.items():
            if key.startswith("download_warning"):
                token = value
                break
        else:
            api_response, content = _extract_gdrive_api_response(response)
            token = "t" if api_response == "Virus scan warning" else None

        if token is not None:
            response = session.get(url, params=dict(params, confirm=token), stream=True)
            api_response, content = _extract_gdrive_api_response(response)

        if api_response == "Quota exceeded":
            raise RuntimeError(
                f"The daily quota of the file {filename} is exceeded and it "
                f"can't be downloaded. This is a limitation of Google Drive "
                f"and can only be overcome by trying again later."
            )

        _save_response_content(content, fpath)

    # In case we deal with an unhandled GDrive API response, the file should be smaller than 10kB and contain only text
    if os.stat(fpath).st_size < 10 * 1024:
        with contextlib.suppress(UnicodeDecodeError), open(fpath) as fh:
            text = fh.read()
            # Regular expression to detect HTML. Copied from https://stackoverflow.com/a/70585604
            if re.search(r"</?\s*[a-z-][^>]*\s*>|(&(?:[\w\d]+|#\d+|#x[a-f\d]+);)", text):
                warnings.warn(
                    f"We detected some HTML elements in the downloaded file. "
                    f"This most likely means that the download triggered an unhandled API response by GDrive. "
                    f"Please report this to torchvision at https://github.com/pytorch/vision/issues including "
                    f"the response:\n\n{text}"
                )

    if md5 and not check_md5(fpath, md5):
        raise RuntimeError(
            f"The MD5 checksum of the download file {fpath} does not match the one on record."
            f"Please delete the file and try again. "
            f"If the issue persists, please report this to torchvision at https://github.com/pytorch/vision/issues."
        )


def _get_google_drive_file_id(url: str) -> Optional[str]:
    parts = urlparse(url)

    if re.match(r"(drive|docs)[.]google[.]com", parts.netloc) is None:
        return None

    match = re.match(r"/file/d/(?P<id>[^/]*)", parts.path)
    if match is None:
        return None

    return match.group("id")


def _urlretrieve(url: str, filename: str, chunk_size: int = 1024 * 32) -> None:
    with urllib.request.urlopen(urllib.request.Request(url, headers={"User-Agent": USER_AGENT})) as response:
        _save_response_content(iter(lambda: response.read(chunk_size), b""), filename, length=response.length)


def _save_response_content(
    content: Iterator[bytes],
    destination: str,
    length: Optional[int] = None,
) -> None:
    with open(destination, "wb") as fh, tqdm(total=length) as pbar:
        for chunk in content:
            # filter out keep-alive new chunks
            if not chunk:
                continue

            fh.write(chunk)
            pbar.update(len(chunk))




def _extract_gdrive_api_response(response, chunk_size: int = 32 * 1024) -> Tuple[bytes, Iterator[bytes]]:
    content = response.iter_content(chunk_size)
    first_chunk = None
    # filter out keep-alive new chunks
    while not first_chunk:
        first_chunk = next(content)
    content = itertools.chain([first_chunk], content)

    try:
        match = re.search("<title>Google Drive - (?P<api_response>.+?)</title>", first_chunk.decode())
        api_response = match["api_response"] if match is not None else None
    except UnicodeDecodeError:
        api_response = None
    return api_response, content

            




# use these by importing them in another script e.g.
# from galaxy_datasets.shared.label_metadata import decals_label_cols 

# The "ortho" versions avoid mixing votes from different campaigns by appending -{campaign} (e.g. -dr5, -gz2) to each question

# includes DECaLS, GZ2, GZ Rings, GZ Cosmic Dawn


def extract_questions_and_label_cols(question_answer_pairs):
    """
    Convenience wrapper to get list of questions and label_cols from a schema.
    Common starting point for analysis, iterating over questions, etc.

    Args:
        question_answer_pairs (dict): e.g. {'smooth-or-featured: ['_smooth, _featured-or-disk, ...], ...}

    Returns:
        list: all questions e.g. [Question('smooth-or-featured'), ...]
        list: label_cols (list of answer strings). See ``label_metadata.py`` for examples.
    """
    questions = list(question_answer_pairs.keys())
    label_cols = [q + answer for q, answers in question_answer_pairs.items() for answer in answers]
    return questions, label_cols



"""
DECALS (GZD-1, GZD-2, GZD-5, GZD-8)
"""

# the schema is slightly different for dr1/2 vs dr5+
# merging answers were changed completely
# bulge sizes were changed from 3 to 5
# bar was changed from yes/no to strong/weak/none
# spiral had 'cant tell' added
decals_dr12_ortho_pairs = {
    'smooth-or-featured-dr12': ['_smooth', '_featured-or-disk', '_artifact'],
    'disk-edge-on-dr12': ['_yes', '_no'],
    'has-spiral-arms-dr12': ['_yes', '_no'],
    'bar-dr12': ['_yes', '_no'],
    'bulge-size-dr12': ['_dominant', '_obvious', '_none'],
    'how-rounded-dr12': ['_completely', '_in-between', '_cigar-shaped'],  # completely was renamed to round
    'edge-on-bulge-dr12': ['_boxy', '_none', '_rounded'],
    'spiral-winding-dr12': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-dr12': ['_1', '_2', '_3', '_4', '_more-than-4'],
    'merging-dr12': ['_neither', '_tidal-debris', '_both', '_merger']
}

decals_dr5_ortho_pairs = {
    'smooth-or-featured-dr5': ['_smooth', '_featured-or-disk', '_artifact'],
    'disk-edge-on-dr5': ['_yes', '_no'],
    'has-spiral-arms-dr5': ['_yes', '_no'],
    'bar-dr5': ['_strong', '_weak', '_no'],
    'bulge-size-dr5': ['_dominant', '_large', '_moderate', '_small', '_none'],
    'how-rounded-dr5': ['_round', '_in-between', '_cigar-shaped'],
    'edge-on-bulge-dr5': ['_boxy', '_none', '_rounded'],
    'spiral-winding-dr5': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-dr5': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell'],
    'merging-dr5': ['_none', '_minor-disturbance', '_major-disturbance', '_merger']
}

decals_dr5_ortho_questions, decals_dr5_ortho_label_cols = extract_questions_and_label_cols(decals_dr5_ortho_pairs)

# exactly the same for dr8. 
decals_dr8_ortho_pairs = decals_dr5_ortho_pairs.copy()
for question, answers in decals_dr8_ortho_pairs.copy().items(): # avoid modifying while looping
    decals_dr8_ortho_pairs[question.replace('-dr5', '-dr8')] = answers
    del decals_dr8_ortho_pairs[question]  # delete the old ones


# I think performance should be best when training on *both*
decals_all_campaigns_ortho_pairs = {}  # big dict including all the pairs from DR1/2, DR5, DR8
decals_all_campaigns_ortho_pairs.update(decals_dr12_ortho_pairs)
decals_all_campaigns_ortho_pairs.update(decals_dr5_ortho_pairs)
decals_all_campaigns_ortho_pairs.update(decals_dr8_ortho_pairs)

# for convenience, extract lists of just the questions and just the label columns
# these can then be imported throughout without needing to apply this function elsewhere
decals_dr12_ortho_questions, decals_dr12_ortho_label_cols = extract_questions_and_label_cols(decals_dr12_ortho_pairs)
decals_dr5_ortho_questions, decals_dr5_ortho_label_cols = extract_questions_and_label_cols(decals_dr5_ortho_pairs)
decals_dr8_ortho_questions, decals_dr8_ortho_label_cols = extract_questions_and_label_cols(decals_dr8_ortho_pairs)
decals_all_campaigns_ortho_questions, decals_all_campaigns_ortho_label_cols = extract_questions_and_label_cols(decals_all_campaigns_ortho_pairs)


# Dict mapping each question (e.g. disk-edge-on) 
# to the answer on which it depends (e.g. smooth-or-featured_featured-or-disk)
decals_ortho_dependencies = {
    # dr12
    'smooth-or-featured-dr12': None,
    'disk-edge-on-dr12': 'smooth-or-featured-dr12_featured-or-disk',
    'has-spiral-arms-dr12': 'disk-edge-on-dr12_no',
    'bar-dr12': 'disk-edge-on-dr12_no',
    'bulge-size-dr12': 'disk-edge-on-dr12_no',
    'how-rounded-dr12': 'smooth-or-featured-dr12_smooth',
    'edge-on-bulge-dr12': 'disk-edge-on-dr12_yes',
    'spiral-winding-dr12': 'has-spiral-arms-dr12_yes',
    'spiral-arm-count-dr12': 'has-spiral-arms-dr12_yes',
    'merging-dr12': None,  # TODO technically should be smooth OR featured, but hard to code - no double dependency support yet
    # dr5
    'smooth-or-featured-dr5': None,  # always asked
    'disk-edge-on-dr5': 'smooth-or-featured-dr5_featured-or-disk',
    'has-spiral-arms-dr5': 'disk-edge-on-dr5_no',
    'bar-dr5': 'disk-edge-on-dr5_no',
    'bulge-size-dr5': 'disk-edge-on-dr5_no',
    'how-rounded-dr5': 'smooth-or-featured-dr5_smooth',
    'edge-on-bulge-dr5': 'disk-edge-on-dr5_yes',
    'spiral-winding-dr5': 'has-spiral-arms-dr5_yes',
    'spiral-arm-count-dr5': 'has-spiral-arms-dr5_yes', # bad naming...
    'merging-dr5': None,
    # dr8 is identical to dr5, just with -dr8
    'smooth-or-featured-dr8': None,
    'disk-edge-on-dr8': 'smooth-or-featured-dr8_featured-or-disk',
    'has-spiral-arms-dr8': 'disk-edge-on-dr8_no',
    'bar-dr8': 'disk-edge-on-dr8_no',
    'bulge-size-dr8': 'disk-edge-on-dr8_no',
    'how-rounded-dr8': 'smooth-or-featured-dr8_smooth',
    'edge-on-bulge-dr8': 'disk-edge-on-dr8_yes',
    'spiral-winding-dr8': 'has-spiral-arms-dr8_yes',
    'spiral-arm-count-dr8': 'has-spiral-arms-dr8_yes',
    'merging-dr8': None,
    }


"""
DESI

Identical to DECalS DR5, DR8, except without the -dr8
Not used for ML, only for user convenience
"""

desi_pairs = decals_dr8_ortho_pairs.copy()
for question, answers in desi_pairs.copy().items(): # avoid modifying while looping
    desi_pairs[question.replace('-dr8', '')] = answers
    del desi_pairs[question]  # delete the old ones

desi_dependencies = {}
for question, dependency in decals_ortho_dependencies.copy().items():
    if '-dr8' in question:
        question_text = question.replace('-dr8', '')
        if dependency is None:
            dependency_text = None
        else:
            dependency_text = dependency.replace('-dr8', '')
        desi_dependencies[question_text] = dependency_text 


"""
Galaxy Zoo 2 (GZ2)
"""

gz2_ortho_pairs = {
    'smooth-or-featured-gz2': ['_smooth', '_featured-or-disk', '_artifact'],
    'disk-edge-on-gz2': ['_yes', '_no'],
    'has-spiral-arms-gz2': ['_yes', '_no'],
    'bar-gz2': ['_yes', '_no'],
    'bulge-size-gz2': ['_dominant', '_obvious', '_just-noticeable', '_no'],
    'something-odd-gz2': ['_yes', '_no'],
    'how-rounded-gz2': ['_round', '_in-between', '_cigar'],
    'bulge-shape-gz2': ['_round', '_boxy', '_no-bulge'],
    'spiral-winding-gz2': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-gz2': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell']
}
gz2_ortho_questions, gz2_ortho_label_cols = extract_questions_and_label_cols(gz2_ortho_pairs)

gz2_ortho_dependencies = {
    'smooth-or-featured-gz2': None,  # always asked
    'disk-edge-on-gz2': 'smooth-or-featured-gz2_featured-or-disk',
    'has-spiral-arms-gz2': 'smooth-or-featured-gz2_featured-or-disk',
    'bar-gz2': 'smooth-or-featured-gz2_featured-or-disk',
    'bulge-size-gz2': 'smooth-or-featured-gz2_featured-or-disk',
    'how-rounded-gz2': 'smooth-or-featured-gz2_smooth',
    'bulge-shape-gz2': 'disk-edge-on-gz2_yes',  # gz2 only
    'edge-on-bulge-gz2': 'disk-edge-on-gz2_yes',
    'spiral-winding-gz2': 'has-spiral-arms-gz2_yes',
    'spiral-arm-count-gz2': 'has-spiral-arms-gz2_yes',
    'something-odd-gz2': None  # actually neglects the artifact branch
}

"""
Galaxy Zoo Rings (GZ Rings)
"""

rings_pairs = {
    'ring': ['_yes', '_no']
}
rings_questions, rings_label_cols = extract_questions_and_label_cols(rings_pairs)

rings_dependencies = {'ring': None }


"""
Galaxy Zoo Hubble (and Euclidized)
"""

# TODO may change features to featured-or-disk
hubble_pairs = {
    'smooth-or-featured': ['_smooth', '_features', '_artifact'],
    'how-rounded': ['_completely', '_in-between', '_cigar-shaped'],
    'clumpy-appearance': ['_yes', '_no'],
    # 'clump-count': ['_1', '_2', '_3', '_4', '_5-plus', '_cant-tell'],
    # disable these for now as I don't support having several but not all answers leading to the same next question
    # 'clump-configuration': ['_straight-line', '_chain', '_cluster-or-irregular', '_spiral'],
    # 'one-clump-brightest': ['_yes', '_no'],
    # 'brightest-clump-central': ['_yes', '_no'],
    'disk-edge-on': ['_yes', '_no'],
    'bulge-shape': ['_rounded', '_boxy', '_none'],
    'bar': ['_yes', '_no'],
    'has-spiral-arms': ['_yes', '_no'],
    'spiral-winding': ['_tight', '_medium', '_loose'],
    'spiral-arm-count': ['_1', '_2', '_3', '_4', '_5-plus', '_cant-tell'],
    'bulge-size': ['_none', '_just-noticeable', '_obvious', '_dominant'],
    'galaxy-symmetrical': ['_yes', '_no'],
    'clumps-embedded-larger-object': ['_yes', '_no']
}
# add -hubble to the end of each question
hubble_ortho_pairs = dict([(key + '-hubble', value) for key, value in hubble_pairs.items()])

# not used here, but may be helpful elsewhere
hubble_ortho_dependencies = {
    'smooth-or-featured-hubble': None,
    'how-rounded-hubble': 'smooth-or-featured-hubble_smooth',
    'clumpy-appearance-hubble': 'smooth-or-featured-hubble_features',
    # 'clump-count-hubble': 'clumpy-appearance-hubble_yes',
    # 'clump-configuration-hubble': ['_straight-line', '_chain', '_cluster-or-irregular', '_spiral'],
    # 'one-clump-brightest-hubble': ['_yes', '_no'],
    # 'brightest-clump-central-hubble': ['_yes', '_no'],
    # ignoring the spiral dashed line, probably rare
    'galaxy-symmetrical-hubble': 'clumpy-appearance-hubble_yes',
    'clumps-embedded-larger-object-hubble': 'clumpy-appearance-hubble_yes',
    'disk-edge-on-hubble': 'clumpy-appearance-hubble_no',
    'bulge-shape-hubble': 'disk-edge-on-hubble_yes',
    'edge-on-bulge-hubble': 'disk-edge-on-hubble_yes',
    'bar-hubble': 'disk-edge-on-hubble_no',
    'has-spiral-arms-hubble': 'disk-edge-on-hubble_no',
    'spiral-winding-hubble': 'disk-edge-on-hubble_no',
    'spiral-arm-count-hubble': 'disk-edge-on-hubble_no',
    'bulge-size-hubble': 'disk-edge-on-hubble_no'
}

hubble_ortho_questions, hubble_ortho_label_cols = extract_questions_and_label_cols(hubble_ortho_pairs)



# similar but simplified version avoiding the questionable clump questions
# Courtesy David O'Ryan - thanks, David!
hubble_v2_pairs = {
    'smooth-or-featured': ['_smooth', '_features', '_artifact'],
    'how-rounded': ['_completely', '_in-between', '_cigar-shaped'],
    'disk-edge-on': ['_yes', '_no'],
    'bulge-shape': ['_rounded', '_boxy', '_none'],
    'bar': ['_yes', '_no'],
    'has-spiral-arms': ['_yes', '_no'],
    'spiral-winding': ['_tight', '_medium', '_loose'],
    'spiral-arm-count': ['_1', '_2', '_3', '_4', '_5-plus', '_cant-tell'],
    'bulge-size': ['_none', '_just-noticeable', '_obvious', '_dominant'],
}
hubble_v2_ortho_pairs = dict([(key + '-hubble', value) for key, value in hubble_v2_pairs.items()])

hubble_v2_ortho_dependencies = {
    'smooth-or-featured-hubble': None,
    'how-rounded-hubble': 'smooth-or-featured-hubble_smooth',
    'disk-edge-on-hubble': 'smooth-or-featured-hubble_features',
    'bulge-shape-hubble': 'disk-edge-on-hubble_yes',
    'edge-on-bulge-hubble': 'disk-edge-on-hubble_yes',
    'bar-hubble': 'disk-edge-on-hubble_no',
    'has-spiral-arms-hubble': 'disk-edge-on-hubble_no',
    'spiral-winding-hubble': 'disk-edge-on-hubble_no',
    'spiral-arm-count-hubble': 'disk-edge-on-hubble_no',
    'bulge-size-hubble': 'disk-edge-on-hubble_no'
}
hubble_v2_ortho_questions, hubble_v2_ortho_label_cols = extract_questions_and_label_cols(hubble_v2_ortho_pairs)

"""
Galaxy Zoo CANDELS
"""


# TODO may change features to featured-or-disk
candels_pairs = {
    'smooth-or-featured': ['_smooth', '_features', '_artifact'],
    'how-rounded': ['_completely', '_in-between', '_cigar-shaped'],
    'clumpy-appearance': ['_yes', '_no'],
    # 'clump-count': ['_1', '_2', '_3', '_4', '_5-plus', '_cant-tell'],
    # disable these for now as I don't support having several but not all answers leading to the same next question
    # 'clump-configuration': ['_straight-line', '_chain', '_cluster-or-irregular', '_spiral'],
    # 'one-clump-brightest': ['_yes', '_no'],
    # 'brightest-clump-central': ['_yes', '_no'],
    # 'galaxy-symmetrical': ['_yes', '_no'],
    # 'clumps-embedded-larger-object': ['_yes', '_no'],
    'disk-edge-on': ['_yes', '_no'],
    'edge-on-bulge': ['_yes', '_no'],
    'bar': ['_yes', '_no'],
    'has-spiral-arms': ['_yes', '_no'],
    'spiral-winding': ['_tight', '_medium', '_loose'],
    'spiral-arm-count': ['_1', '_2', '_3', '_4', '_5-plus', '_cant-tell'],
    'bulge-size': ['_none', '_obvious', '_dominant'],
    'merging': ['_merger', '_tidal-debris', '_both', '_neither']
}
# add -candels to the end of each question
candels_ortho_pairs = dict([(key + '-candels', value) for key, value in candels_pairs.items()])

# not used here, but may be helpful elsewhere
candels_ortho_dependencies = {
    'smooth-or-featured-candels': None,
    'how-rounded-candels': 'smooth-or-featured-candels_smooth',
    'clumpy-appearance-candels': 'smooth-or-featured-candels_features',
    # 'clump-count-candels': 'clumpy-appearance-candels_yes',
    # 'clump-configuration-candels': ['_straight-line', '_chain', '_cluster-or-irregular', '_spiral'],
    # 'one-clump-brightest-candels': ['_yes', '_no'],
    # 'brightest-clump-central-candels': ['_yes', '_no'],
    # 'galaxy-symmetrical-candels': ['_yes', '_no'],
    # 'clumps-embedded-larger-object-candels': ['_yes', '_no'],
    'disk-edge-on-candels': 'clumpy-appearance-candels_no',
    'edge-on-bulge-candels': 'disk-edge-on-candels_yes',
    'bar-candels': 'disk-edge-on-candels_no',
    'has-spiral-arms-candels': 'disk-edge-on-candels_no',
    'spiral-winding-candels': 'disk-edge-on-candels_no',
    'spiral-arm-count-candels': 'disk-edge-on-candels_no',
    'bulge-size-candels': 'disk-edge-on-candels_no',
    'merging-candels': None
}

candels_ortho_questions, candels_ortho_label_cols = extract_questions_and_label_cols(candels_ortho_pairs)


"""
Galaxy Zoo Cosmic Dawn (HSC)

Not yet created in galaxy-datasets
"""


cosmic_dawn_ortho_pairs = {
    'smooth-or-featured-cd': ['_smooth', '_featured-or-disk', '_problem'],  # renamed from artifact
    'disk-edge-on-cd': ['_yes', '_no'],
    'has-spiral-arms-cd': ['_yes', '_no'],
    'bar-cd': ['_strong', '_weak', '_no'],
    'bulge-size-cd': ['_dominant', '_large', '_moderate', '_small', '_none'],
    'how-rounded-cd': ['_round', '_in-between', '_cigar-shaped'],
    'edge-on-bulge-cd': ['_boxy', '_none', '_rounded'],
    'spiral-winding-cd': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-cd': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell'],
    'merging-cd': ['_none', '_minor-disturbance', '_major-disturbance', '_merger'],
    'clumps-cd': ['_yes', '_no'],
    'problem-cd': ['_star', '_artifact', '_zoom'],
    'artifact-cd': ['_satellite', '_scattered', '_diffraction', '_ray', '_saturation', '_other']
}

cosmic_dawn_ortho_dependencies = {
    'smooth-or-featured-cd': None,
    'problem-cd': 'smooth-or-featured-cd_problem',  # newly added (new problem branch)
    'artifact-cd': 'problem-cd_artifact',  # newly added (new problem branch)
    'disk-edge-on-cd': 'smooth-or-featured-cd_featured-or-disk',
    'clumps-cd': 'disk-edge-on-cd_no',  # newly added (featured branch)
    'has-spiral-arms-cd': 'disk-edge-on-cd_no',
    'bar-cd': 'disk-edge-on-cd_no',
    'bulge-size-cd': 'disk-edge-on-cd_no',
    'how-rounded-cd': 'smooth-or-featured-cd_smooth',
    'edge-on-bulge-cd': 'disk-edge-on-cd_yes',
    'spiral-winding-cd': 'has-spiral-arms-cd_yes',
    'spiral-arm-count-cd': 'has-spiral-arms-cd_yes',
    'merging-cd': None  # technically, should be smooth OR featured, but hard to code
}

cosmic_dawn_ortho_questions, cosmic_dawn_ortho_label_cols = extract_questions_and_label_cols(cosmic_dawn_ortho_pairs)

def change_suffix(old_pairs, old_dependancies, old_suffix, new_suffix):
    pairs = old_pairs.copy()
    dependencies = {}

    for question, answers in pairs.copy().items(): # avoid modifying while looping
        pairs[question.replace(old_suffix, new_suffix)] = answers
        del pairs[question]  # delete the old ones

    for question, dependency in old_dependancies.copy().items():
        question_text = question.replace(old_suffix, new_suffix)
        if dependency is None:
            dependency_text = None
        else:
            dependency_text = dependency.replace(old_suffix, new_suffix)
        dependencies[question_text] = dependency_text 

    return pairs, dependencies


cosmic_dawn_pairs, cosmic_dawn_dependencies = change_suffix(cosmic_dawn_ortho_pairs, cosmic_dawn_ortho_dependencies, old_suffix='-cd', new_suffix='')

# jwst is exactly the same except with -jwst instead of -cd
# (has diffraction spikes under rare features, but we don't train on that anyway)
jwst_ortho_pairs, jwst_ortho_dependencies = change_suffix(cosmic_dawn_ortho_pairs, cosmic_dawn_ortho_dependencies, old_suffix='-cd', new_suffix='-jwst')
"""
UKIDSS (almost the same as GZ2)

"""

ukidss_ortho_pairs = {
    'smooth-or-featured-ukidss': ['_smooth', '_featured-or-disk', '_artifact'],
    'disk-edge-on-ukidss': ['_yes', '_no'],
    'has-spiral-arms-ukidss': ['_yes', '_no'],
    'bar-ukidss': ['_yes', '_no'],
    'bulge-size-ukidss': ['_dominant', '_obvious', '_just-noticeable', '_no'],
    'something-odd-ukidss': ['_yes', '_no'],
    'how-rounded-ukidss': ['_round', '_in-between', '_cigar'],
    'bulge-shape-ukidss': ['_round', '_boxy', '_no-bulge'],
    'spiral-winding-ukidss': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-ukidss': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell']
}
ukidss_ortho_questions, ukidss_ortho_label_cols = extract_questions_and_label_cols(ukidss_ortho_pairs)

ukidss_ortho_dependencies = {
    'smooth-or-featured-ukidss': None,  # always asked
    'disk-edge-on-ukidss': 'smooth-or-featured-ukidss_featured-or-disk',
    'has-spiral-arms-ukidss': 'smooth-or-featured-ukidss_featured-or-disk',
    'bar-ukidss': 'smooth-or-featured-ukidss_featured-or-disk',
    'bulge-size-ukidss': 'smooth-or-featured-ukidss_featured-or-disk',
    'how-rounded-ukidss': 'smooth-or-featured-ukidss_smooth',
    'bulge-shape-ukidss': 'disk-edge-on-ukidss_yes',  # ukidss only
    'edge-on-bulge-ukidss': 'disk-edge-on-ukidss_yes',
    'spiral-winding-ukidss': 'has-spiral-arms-ukidss_yes',
    'spiral-arm-count-ukidss': 'has-spiral-arms-ukidss_yes',
    'something-odd-ukidss': None  # actually neglects the artifact branch
}


# same as cosmic dawn (except ghosts) for now
euclid_ortho_pairs = {
    'smooth-or-featured-euclid': ['_smooth', '_featured-or-disk', '_problem'],  # renamed from artifact
    'disk-edge-on-euclid': ['_yes', '_no'],
    'has-spiral-arms-euclid': ['_yes', '_no'],
    'bar-euclid': ['_strong', '_weak', '_no'],
    'bulge-size-euclid': ['_dominant', '_large', '_moderate', '_small', '_none'],
    'how-rounded-euclid': ['_round', '_in-between', '_cigar-shaped'],
    'edge-on-bulge-euclid': ['_boxy', '_none', '_rounded'],
    'spiral-winding-euclid': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-euclid': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell'],
    'merging-euclid': ['_none', '_minor-disturbance', '_major-disturbance', '_merger'],
    'clumps-euclid': ['_yes', '_no'],
    'problem-euclid': ['_star', '_artifact', '_zoom'],
    'artifact-euclid': ['_satellite', '_scattered', '_diffraction', '_ray', '_saturation', '_other', '_ghost']
}

euclid_ortho_dependencies = {
    'smooth-or-featured-euclid': None,
    'problem-euclid': 'smooth-or-featured-euclid_problem',  # newly added (new problem branch)
    'artifact-euclid': 'problem-euclid_artifact',  # newly added (new problem branch)
    'disk-edge-on-euclid': 'smooth-or-featured-euclid_featured-or-disk',
    'clumps-euclid': 'disk-edge-on-euclid_no',  # newly added (featured branch)
    'has-spiral-arms-euclid': 'disk-edge-on-euclid_no',
    'bar-euclid': 'disk-edge-on-euclid_no',
    'bulge-size-euclid': 'disk-edge-on-euclid_no',
    'how-rounded-euclid': 'smooth-or-featured-euclid_smooth',
    'edge-on-bulge-euclid': 'disk-edge-on-euclid_yes',
    'spiral-winding-euclid': 'has-spiral-arms-euclid_yes',
    'spiral-arm-count-euclid': 'has-spiral-arms-euclid_yes',
    'merging-euclid': None  # technically, should be smooth OR featured, but hard to code
}

euclid_ortho_questions, euclid_ortho_label_cols = extract_questions_and_label_cols(euclid_ortho_pairs)

euclid_pairs, euclid_dependencies = change_suffix(euclid_ortho_pairs, euclid_ortho_dependencies, old_suffix='-euclid', new_suffix='')
euclid_questions, euclid_label_cols = extract_questions_and_label_cols(euclid_pairs)




def get_gz_evo_v1_metadata(internal):

    question_answer_pairs = {}
    question_answer_pairs.update(decals_all_campaigns_ortho_pairs)
    question_answer_pairs.update(gz2_ortho_pairs)
    question_answer_pairs.update(candels_ortho_pairs)
    question_answer_pairs.update(hubble_ortho_pairs)
    if internal:
        question_answer_pairs.update(cosmic_dawn_ortho_pairs)

    dependencies = {}
    dependencies.update(decals_ortho_dependencies)
    dependencies.update(gz2_ortho_dependencies)
    dependencies.update(candels_ortho_dependencies)
    dependencies.update(hubble_ortho_dependencies)
    if internal:
        dependencies.update(cosmic_dawn_ortho_dependencies)

    label_cols = \
        decals_all_campaigns_ortho_label_cols + \
        gz2_ortho_label_cols + \
        candels_ortho_label_cols + \
        hubble_ortho_label_cols
    if internal:
        label_cols += cosmic_dawn_ortho_label_cols

    return label_cols, question_answer_pairs, dependencies

gz_evo_v1_label_cols, gz_evo_v1_pairs, gz_evo_v1_dependencies = get_gz_evo_v1_metadata(internal=True)
gz_evo_v1_public_label_cols, gz_evo_v1_public_pairs, gz_evo_v1_public_dependencies = get_gz_evo_v1_metadata(internal=False)


""" same again for V2, adding Hubble V2 and Euclid"""
def get_gz_evo_v2_metadata(internal):

    question_answer_pairs = {}
    question_answer_pairs.update(decals_all_campaigns_ortho_pairs)
    question_answer_pairs.update(gz2_ortho_pairs)
    question_answer_pairs.update(candels_ortho_pairs)
    question_answer_pairs.update(hubble_v2_ortho_pairs)
    if internal:
        question_answer_pairs.update(cosmic_dawn_ortho_pairs)
        question_answer_pairs.update(euclid_ortho_pairs)

    dependencies = {}
    dependencies.update(decals_ortho_dependencies)
    dependencies.update(gz2_ortho_dependencies)
    dependencies.update(candels_ortho_dependencies)
    dependencies.update(hubble_v2_ortho_dependencies)
    if internal:
        dependencies.update(cosmic_dawn_ortho_dependencies)
        dependencies.update(euclid_ortho_dependencies)

    label_cols = \
        decals_all_campaigns_ortho_label_cols + \
        gz2_ortho_label_cols + \
        candels_ortho_label_cols + \
        hubble_v2_ortho_label_cols
    if internal:
        label_cols += cosmic_dawn_ortho_label_cols
        label_cols += euclid_ortho_label_cols

    return label_cols, question_answer_pairs, dependencies

gz_evo_v2_label_cols, gz_evo_v2_pairs, gz_evo_v2_dependencies = get_gz_evo_v2_metadata(internal=True)
gz_evo_v2_public_label_cols, gz_evo_v2_public_pairs, gz_evo_v2_public_dependencies = get_gz_evo_v2_metadata(internal=False)

# JWST CEERS 
jwst_ortho_pairs = {
    'smooth-or-featured-jwst': ['_smooth', '_featured-or-disk', '_star-artifact-zoom'],
    'disk-edge-on-jwst': ['_yes', '_no'],
    'has-spiral-arms-jwst': ['_yes', '_no'],
    'bar-jwst': ['_strong', '_weak', '_no'],
    'bulge-size-jwst': ['_dominant', '_large', '_moderate', '_small', '_none'],
    'how-rounded-jwst': ['_round', '_in-between', '_cigar-shaped'],
    'edge-on-bulge-jwst': ['_boxy', '_none', '_rounded'],
    'spiral-winding-jwst': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-jwst': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell'],
    'clumps-jwst': ['_yes', '_no'],
    'merging-jwst': ['_none', '_minor-disturbance', '_major-disturbance', '_merger'],
    'problem-jwst': ['_star', '_artifact', '_bad-zoom']
}
jwst_ortho_questions, jwst_ortho_label_cols = extract_questions_and_label_cols(jwst_ortho_pairs)

jwst_ortho_dependencies = {
    'smooth-or-featured-jwst': None,  # always asked
    'disk-edge-on-jwst': 'smooth-or-featured-jwst_featured-or-disk',
    'has-spiral-arms-jwst': 'disk-edge-on-jwst_no',
    'bar-jwst': 'disk-edge-on-jwst_no',
    'bulge-size-jwst': 'disk-edge-on-jwst_no',
    'how-rounded-jwst': 'smooth-or-featured-jwst_smooth',
    'edge-on-bulge-jwst': 'disk-edge-on-jwst_yes',
    'spiral-winding-jwst': 'has-spiral-arms-jwst_yes',
    'spiral-arm-count-jwst': 'has-spiral-arms-jwst_yes', # bad naming...
    'merging-jwst': None,  # ignores artifact,
    'clumps-jwst': None,  # ignores artifact
    'problem-jwst': 'smooth-or-featured-jwst_star-artifact-zoom'
}

# JWST COSMOS
jwst_cosmos_ortho_pairs = {
    'smooth-or-featured-jwst': ['_smooth', '_featured-or-disk', '_star-artifact-zoom'],
    'disk-edge-on-jwst': ['_yes', '_no'],
    'has-spiral-arms-jwst': ['_yes', '_no'],
    'bar-jwst': ['_strong', '_weak', '_no'],
    'bulge-size-jwst': ['_dominant', '_large', '_moderate', '_small', '_none'],
    'edge-on-bulge-jwst': ['_boxy', '_none', '_rounded'],
    'spiral-winding-jwst': ['_tight', '_medium', '_loose'],
    'spiral-arm-count-jwst': ['_1', '_2', '_3', '_4', '_more-than-4', '_cant-tell'],
    'clumps-jwst': ['_yes', '_no'],
    'merging-jwst': ['_none', '_minor-disturbance', '_major-disturbance', '_merger'],
    'problem-jwst': ['_star', '_artifact', '_bad-zoom']
}
jwst_cosmos_ortho_questions, jwst_cosmos_ortho_label_cols = extract_questions_and_label_cols(jwst_cosmos_ortho_pairs)

jwst_cosmos_ortho_dependencies = {
    'smooth-or-featured-jwst': None,  # always asked
    'disk-edge-on-jwst': 'smooth-or-featured-jwst_featured-or-disk',
    'has-spiral-arms-jwst': 'disk-edge-on-jwst_no',
    'bar-jwst': 'disk-edge-on-jwst_no',
    'bulge-size-jwst': 'disk-edge-on-jwst_no',
    'edge-on-bulge-jwst': 'disk-edge-on-jwst_yes',
    'spiral-winding-jwst': 'has-spiral-arms-jwst_yes',
    'spiral-arm-count-jwst': 'has-spiral-arms-jwst_yes',
    'merging-jwst': None,  # ignores artifact,
    'clumps-jwst': None,  # ignores artifact
    'problem-jwst': 'smooth-or-featured-jwst_star-artifact-zoom'
}



import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata
from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
if not INTERNAL_URLS_EXIST:
    raise FileNotFoundError
from galaxy_datasets.shared import internal_urls


def gz_jwst(root, train, download):
    logging.info('Setting up gz_jwst dataset')
    resources = [
        (internal_urls.gz_jwst_train_catalog, '69e689017e8317abce6b2715382746ae'),  # train catalog
        (internal_urls.gz_jwst_test_catalog, 'c23807b06708bbdc5fdaefa284a96094'),  # test catalog
        (internal_urls.gz_jwst_images, '28ed102a4bcf4bb6693caeebd76a9c44')  # the images
    ]
    images_to_spotcheck = [
        '88113560.jpg'
    ]

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    label_cols = label_metadata.jwst_ortho_label_cols
    useful_columns = label_cols + ['filename', 'subject_id', 'id_str']
    # useful_columns = None
    if train:
        train_catalog_loc = os.path.join(root, 'jwst_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'jwst_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    # slight hack as I used subject_id as the filename semi-accidentally
    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, str(x['subject_id'])+'.jpg'), axis=1)

    logging.info('gz_jwst dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    gz_jwst(root='/home/walml/repos/galaxy-datasets/roots/gz_jwst_debug', train=True, download=False)




import albumentations as A

from galaxy_datasets.transforms_albumentations import ToGray, RemoveAlpha

def astroaugmentation_transforms(
    resize_size: int,
    shift_limit: float,
    scale_limit: float,
    rotate_limit: float,
    p_channelwise_dropout: float,
    p_elastic=0,
    elastic_alpha=1,
    elastic_sigma=1,
    channelwise_dropout_max_fraction=0.2,
    channelwise_dropout_min_length=10,
    channelwise_dropout_max_holes=100,
    pytorch_greyscale=False,
) -> A.Compose:
    """
    Astrophysically-inspired image transforms from AstroAugmentations (Bowles in prep.)
    Intended to mimic common distortions to optical telescope images.
    
    Sequentially:
    - ElasticTransform (optional) from https://ieeexplore.ieee.org/document/1227801
    - ChannelWiseDropout (good for Decals) from AstroAugmentations
    - ShiftScaleRotate instead of RandomResizedCrop for more control
    - Flip along x or y axis

    Args:
        resize_size (int): resolution
        shift_limit (float): max relative shift factor. 0.3 can have the galaxy pretty much
                            at the edge *without scaling*
        scale_limit (float):  max relative max zoom *in* factor
        rotate_limit (float): max rotate angle in degrees
        p_channelwise_dropout (float): _description_
        p_elastic (int, optional): how much to scale the random field. Defaults to 0.
        elastic_alpha (int, optional): std. of the Gauss. kernel blurring the random field. Defaults to 1.
        elastic_sigma (int, optional):  max area fraction to be affected. Defaults to 1.
        channelwise_dropout_max_fraction (float, optional): _description_. Defaults to 0.2.
        channelwise_dropout_min_length (int, optional):  minimum length of the hole in pixels. Defaults to 10. May be horizontal or vertical (width or height)
        channelwise_dropout_max_holes (int, optional): _description_. Defaults to 100.
        pytorch_greyscale (bool, optional): _description_. Defaults to False.

    Returns:
        A.Compose: _description_
    """
    # wrapped in Try/Except to avoid making AstroAugmentations a package requirement for only this optional transform
    try:
        from AstroAugmentations import image_domain  # type: ignore
    except ImportError:
        raise ImportError(
            'Trying to use astroaugmentation_transforms but AstroAugmentations is not installed\n \
            Please install via git. See instructions at https://github.com/mb010/AstroAugmentations#quick-start'
        )
    from albumentations.pytorch import ToTensorV2  # also required pytorch

    transforms_to_apply = [
        A.Lambda(name="RemoveAlpha", image=RemoveAlpha(), always_apply=True)
    ]

    if pytorch_greyscale:
        transforms_to_apply += [
            A.Lambda(
                name="ToGray", image=ToGray(reduce_channels=True), always_apply=True
            )
        ]

    transforms_to_apply += [
        A.LongestMaxSize(max_size=resize_size),
        A.ElasticTransform(
            alpha=elastic_alpha,
            sigma=elastic_sigma,
            alpha_affine=0,
            interpolation=1,
            border_mode=1,
            value=0,
            p=p_elastic,
        ),
        A.Lambda(
            name="MissingData",
            image=image_domain.optical.ChannelWiseDropout(
                max_fraction=channelwise_dropout_max_fraction,
                # max is up to channelwise_dropout_max_fraction
                min_width=channelwise_dropout_min_length,
                min_height=channelwise_dropout_min_length,
                max_holes=channelwise_dropout_max_holes,
                channelwise_application=True,
            ),
            p=p_channelwise_dropout,
        ),
        A.ShiftScaleRotate(
            shift_limit=shift_limit,
            scale_limit=(0, scale_limit), # type: ignore
            rotate_limit=rotate_limit, # type: ignore
            interpolation=2,
            border_mode=0,
            p=1,
        ),
        # TODO maybe add sersic/gaussian
        A.Flip(p=0.5),
        ToTensorV2(),
    ]

    return A.Compose(transforms_to_apply)




import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata
from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
if not INTERNAL_URLS_EXIST:
    raise FileNotFoundError
from galaxy_datasets.shared import internal_urls


def gz_ukidss(root, train, download):
    logging.info('Setting up gz_ukidss dataset')
    resources = [
        (internal_urls.gz_ukidss_train_catalog, '8d84de1a1f33556f83470ce0c22c979b'),  # train catalog
        (internal_urls.gz_ukidss_test_catalog, 'f64b850e941b09546df5aca84b708fb4'),  # test catalog
        (internal_urls.gz_ukidss_images, 'd4586af013b1543753346544b4703711')  # the images
    ]
    images_to_spotcheck = [
        '524482bc3ae74054bf009/524482bc3ae74054bf009a00.jpg'
    ]

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    label_cols = label_metadata.ukidss_ortho_label_cols
    useful_columns = label_cols + ['filename', 'subfolder']
    # useful_columns = None
    if train:
        train_catalog_loc = os.path.join(root, 'ukidss_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'ukidss_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    logging.info('gz_ukidss dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    gz_ukidss(root='/home/walml/repos/galaxy-datasets/roots/gz_ukidss_debug', train=True, download=True)



# example transforms used for custom/specialized datasets
# e.g. handling Zooniverse grid subjects for GZ JWST COSMOS

import torch
import torchvision.transforms.v2 as T


class FixedCrop(torch.nn.Module):

    def __init__(self, lower_left_x, lower_left_y, upper_right_x, upper_right_y):
        super().__init__()
        # PIL format crop
        self.lower_left_x = lower_left_x
        self.lower_left_y = lower_left_y
        self.upper_right_x = upper_right_x
        self.upper_right_y = upper_right_y
        # pretty ugly, but I'm actually not sure how to only slice the last two dimensions without doing a reshape
        self.crop_hw = T.Lambda(lambda img: img[self.lower_left_y:self.upper_right_y, self.lower_left_x:self.upper_right_x])
        self.crop_chw = T.Lambda(lambda img: img[:, self.lower_left_y:self.upper_right_y, self.lower_left_x:self.upper_right_x])
        self.crop_nchw = T.Lambda(lambda img: img[:, :, self.lower_left_y:self.upper_right_y, self.lower_left_x:self.upper_right_x])

    def forward(self, img):
        if len(img.shape) == 2:
            # grayscale
            return self.crop_hw(img)    
        elif len(img.shape) == 3:
            # RGB
            return self.crop_chw(img)
        elif len(img.shape) == 4:
            # NCHW
            return self.crop_nchw(img)
        else:
            raise ValueError(f'Expected 2D, 3D or 4D tensor, got {len(img.shape)}D tensor instead. Shape: {img.shape}')

if __name__ == '__main__':

    from PIL import Image
    import matplotlib.pyplot as plt

    loc = '/home/walml/repos/galaxy-datasets/tests/examples/1_A1-composite.png'
    img = Image.open(loc)

    print(img.size)

    transforms = [
        T.PILToTensor(),
        FixedCrop(30, 30, 750, 750)
    ]
    transform = T.Compose(transforms)
    img = transform(img)
    print(img.shape)

    plt.imshow(img.permute(1, 2, 0))
    plt.axis('off')
    plt.show()


import logging

from omegaconf import DictConfig

import torch
import torchvision.transforms.v2 as T
from torchvision.transforms import InterpolationMode

from galaxy_datasets.transforms_custom_torchvision import FixedCrop

class GalaxyViewTransform():
    # most powerful and most general transform options

    def __init__(
        self,
        cfg: DictConfig
    ):
        """
        Transforms expect to act on channel-first tensor. 
        Probably uint8 0-255 for speed, but not necessary.

        Transforms return a channel-first float32 tensor in the range (0, 1)

        Args:
            cfg (DictConfig): options for common torchvision transforms
        """
        logging.info(f'Using view transform config: {cfg}')
        
        interpolation = InterpolationMode[cfg.interpolation_method.upper()]
        antialias = True
        
        transform = []

        if cfg.pil_to_tensor:  # galaxydataset loads as PIL image, tensor is faster to work with
            transform.append(T.PILToTensor())
        
        if cfg.greyscale:
            transform.append(T.Grayscale(num_output_channels=1))

        # transform += [T.ToImage()]  # explicit but not needed, just tells torchvision it's an image (assumed anway)
        # disabled for beluga
        # 'torchvision.transforms.v2' has no attribute 'ToImage'
        # requires 0.16.0, beluga has 0.15.2

        if cfg.fixed_crop:
            # fixed crop with specified corners, useful to pick a single image from Zooniverse grid subjects
            transform.append(FixedCrop(
                **cfg.fixed_crop  # lower_left_x, lower_left_y, upper_right_x, upper_right_y
            ))
            logging.info(f'Using FixedCrop, {cfg.fixed_crop}')

        # shear/perspective transforms happen before any cropping
        if cfg.random_affine:  # will use this for rotation and off-center zoom/crop
            transform.append(T.RandomAffine(
                **cfg.random_affine, interpolation=interpolation
            ))
            transform.append(T.Resize(cfg.output_size, interpolation=interpolation, antialias=antialias))
            logging.info(f'Using RandomAffine with safety crop, {cfg.random_affine}')
            transform.append(T.CenterCrop(cfg.output_size))  # T.Resize is a touch imprecise e.g. (224, 227) after resizing. Do a final center crop for safety because torch requires exact shapes.

        else:  # maybe do perspective/crop, maybe do rotation, maybe both
            logging.info('Not using RandomAffine')

            if cfg.random_perspective:
                logging.info(f'Using RandomPerspective, {cfg.random_perspective}')
                transform.append(T.RandomPerspective(**cfg.random_perspective, interpolation=interpolation))
                transform.append(T.CenterCrop(size=cfg.output_size))
            else: # no affine transform, no perspective shift, so safe to apply simple center or random crops
                if cfg.center_crop:
                    transform.append(T.CenterCrop(
                            size=cfg.output_size
                        ))
                if cfg.random_resized_crop:
                    transform.append(T.RandomResizedCrop(
                                size=cfg.output_size,
                                interpolation=interpolation,
                                antialias=antialias,
                                **cfg.random_resized_crop  # scale only
                            ))
                    # transform.append(T.CenterCrop(cfg.output_size))  # T.RandomResizedCrop is also a touch imprecise e.g. (224, 227) after resizing.

            if cfg.rotation_prob > 0:
                    logging.info(f'Using RandomRotation, {cfg.rotation_prob}')
                    # I ASSUME this doesn't change the shape?
                    transform.append(T.RandomRotation(degrees=90, interpolation=interpolation, p=cfg.rotation_prob))


        # flip and rotate. safe after any geometric transforms.
        transform.append(T.RandomHorizontalFlip(p=cfg.flip_prob))
        transform.append(T.RandomVerticalFlip(p=cfg.flip_prob))

        if cfg.color_jitter_prob > 0.:
            transform.append(T.RandomApply(
                        [
                            T.ColorJitter(
                                **cfg.color_jitter
                            )
                        ],
                        p=cfg.color_jitter_prob,
                    ))

        # ugly unpack/tuple cast because T checks for list or tuple, but mine is Omega ListConfig
        if cfg.erase_iterations > 0:
            for _ in range(cfg.erase_iterations):
                scale = tuple(cfg.random_erasing.scale)
                ratio = tuple(cfg.random_erasing.ratio)
                transform.append(T.RandomErasing(scale=scale, ratio=ratio, p=cfg.random_erasing.p))
                
        if cfg.posterize:
            transform.append(T.RandomPosterize(**cfg.posterize))

        if cfg.elastic_prob > 0:
            # see-through-water 
            transform.append(
                T.RandomApply([
                    # should probably not change this 
                    T.ElasticTransform(
                        **cfg.elastic
                         # alpha=magnitude i.e. how much to displace
                         # sigma=smoothness i.e. size of the ripples. Should be about 10+ to keep a distorted galaxy with same morphology. Gets slower fast.
                    )],
                p=cfg.elastic_prob)
            )

        # finally, shift to 0-1 float32 before hitting model etc
        transform.append(T.ToDtype(torch.float32, scale=True))
        # transform.append(T.ConvertImageDtype(torch.float32))

        self.transform = T.Compose(transform)


    def __call__(self, image) -> torch.Tensor:

        return self.transform(image)


def default_view_config():
    # for debugging
     return DictConfig(dict(
        pil_to_tensor=True,
        fixed_crop=False,
        output_size=224,
        greyscale=False,
        interpolation_method='bilinear',
        random_affine=dict(degrees=90, translate=(0.1, 0.1), scale=(1.2, 1.4), shear=(0,20,0,20)),
        random_perspective=False,
        center_crop=False,
        random_resized_crop=False,
        flip_prob=0.5,
        rotation_prob=0.,
        color_jitter_prob=0.,
        # color_jitter=dict(
        #     brightness=.5,
        #     contrast=.3,
        #     saturation=.5,
        #     hue=None
        # ),
        erase_iterations=5,
        random_erasing=dict(p=1., scale=[0.002, 0.007], ratio=[0.5, 2.]),
        posterize=False,
        elastic_prob=0.
        # elastic=dict(alpha=100, sigma=10.)
    ))

def minimal_view_config():
    return DictConfig(dict(
        pil_to_tensor=True,
        fixed_crop=False,
        output_size=224,
        greyscale=False,
        interpolation_method='bilinear',
        # no rotation, no translate, less aggressive crop, no shear
        random_affine=dict(degrees=0, translate=None, scale=(1.2, 1.2), shear=None),
        random_perspective=False,
        center_crop=False,
        random_resized_crop=False,
        flip_prob=0.,
        rotation_prob=0.,
        color_jitter_prob=0.,
        erase_iterations=0,
        posterize=False,
        elastic_prob=0.
    ))

def fast_view_config():
    return DictConfig(dict(
        pil_to_tensor=True,
        fixed_crop=False,
        output_size=224,
        greyscale=False,
        interpolation_method='bilinear',
        random_affine=False,
        random_perspective=False,
        center_crop=True,  # only transform
        random_resized_crop=False,
        flip_prob=0.,
        rotation_prob=0.,
        color_jitter_prob=0.,
        erase_iterations=0,
        posterize=False,
        elastic_prob=0.
    ))

# for now, will deprecate
from galaxy_datasets.transforms_albumentations import minimal_transforms, fast_transforms, default_transforms, base_transforms



import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata
from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
if not INTERNAL_URLS_EXIST:
    raise FileNotFoundError
from galaxy_datasets.shared import internal_urls


def gz_h2o(root, train, download):
    logging.info('Setting up gz_h2o dataset')
    resources = [
        (internal_urls.gz_h2o_train_catalog, '4fa5003cdc1c3beb137be69c06bf796c'),  # train catalog
        (internal_urls.gz_h2o_test_catalog, 'ab4f58091f1c0dee01d9c884c4e9311e'),  # test catalog
        (internal_urls.gz_h2o_images, '805ef596b4271288607ddda5a6acf60a')  # the images
    ]
    images_to_spotcheck = [
        'lz_595/lz_59532_cutout.jpg'
    ]

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    label_cols = label_metadata.cosmic_dawn_ortho_label_cols
    useful_columns = label_cols + ['filename', 'subfolder', 'ra', 'dec']
    # useful_columns = None
    if train:
        train_catalog_loc = os.path.join(root, 'h2o_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'h2o_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    logging.info('gz_h2o dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    gz_h2o(root='/home/walml/repos/galaxy-datasets/roots/gz_h2o_debug', train=True, download=True)




import numpy as np
import albumentations as A
from albumentations.pytorch import ToTensorV2

"""And below, the older transform options"""

# deprecated in favour of torchvision and will soon be removed
def default_transforms(
    crop_scale_bounds=(0.7, 0.8),
    crop_ratio_bounds=(0.9, 1.1),
    initial_center_crop=None,
    resize_after_crop=224, 
    pytorch_greyscale=False,
    to_float=True,  # set to True when loading images directly, False via webdatasets (which normalizes to 0-1 on decode)
    to_tensor=True
    ) -> A.Compose:

    transforms_to_apply = base_transforms(pytorch_greyscale)

    if initial_center_crop:
        transforms_to_apply += [
            A.CenterCrop(
                height=initial_center_crop,  # initial crop
                width=initial_center_crop,
                # always_apply=True
            )
        ]

    transforms_to_apply += [
        # A.ToFloat(),
        # anything outside of the original image is set to 0.
        A.Rotate(limit=180, interpolation=1,
                    always_apply=True, border_mode=0, value=0),
        A.RandomResizedCrop(
            size=(resize_after_crop, resize_after_crop), # after crop resize
            scale=crop_scale_bounds,  # crop factor
            ratio=crop_ratio_bounds,  # crop aspect ratio
            interpolation=1,  # This is "INTER_LINEAR" == BILINEAR interpolation. See: https://docs.opencv.org/3.4/da/d54/group__imgproc__transform.html
            # always_apply=True
        ),  # new aspect ratio
        A.VerticalFlip(p=0.5)
    ]
    if to_float:
        transforms_to_apply += [A.ToFloat(max_value=255.0, always_apply=True)]

    if to_tensor:
        transforms_to_apply += [ToTensorV2(always_apply=True)]

    return A.Compose(transforms_to_apply)



def minimal_transforms(
    resize_after_crop=224, 
    pytorch_greyscale=False
    ) -> A.Compose:
    # for testing how much augmentations slow training / picking CPU to allocate
    transforms_to_apply = base_transforms(pytorch_greyscale)
    transforms_to_apply += [
        A.CenterCrop(
            height=resize_after_crop,
            width=resize_after_crop,
            always_apply=True
        )
    ]
    return A.Compose(transforms_to_apply)


def fast_transforms(
    pytorch_greyscale=False,
    resize_after_crop=224
    ):
    # middle ground between default and minimal transforms
    # faster than default because we avoid interpolation
    # better than minimal because we have some rotation/flip, and use random (non-central) crop
    # should only be used for proper training if you already resize the images
    # such that the resize_after_crop FoV makes sense (as this is just cropping)
    # for 0.75x=224, x=300, so save at 300x300 pixels!
    assert resize_after_crop == 224  # check user isn't attempting to change this
    transforms_to_apply = base_transforms(pytorch_greyscale)
    transforms_to_apply += [
    #     A.RandomCrop(
    #         height=resize_after_crop,
    #         width=resize_after_crop,
    #         always_apply=True
    #     ),
        A.Flip(),
        A.RandomRotate90()
    ]
    return A.Compose(transforms_to_apply)


def base_transforms(pytorch_greyscale):
    if pytorch_greyscale:
        return [
            A.Lambda(
                name="ToGray", image=ToGray(reduce_channels=True) #, always_apply=True
            )
        ]
    else:
       return [
            A.Lambda(name="RemoveAlpha", image=RemoveAlpha()) #, always_apply=True)
        ]
        


# albumentations version of GrayscaleUnweighted
class ToGray():

    def __init__(self, reduce_channels=False):
        self.reduce_channels = reduce_channels

    def forward(self, img):
        if len(img.shape) == 2:  # saved to disk as greyscale already, with no channel
            img = np.expand_dims(img, axis=2) # add channel=1 dimension
        # print(img.shape)
        if self.reduce_channels:
            return img.mean(axis=2, keepdims=True)
        else:
            return img.mean(axis=2, keepdims=True).repeat(3, axis=2)

    def __call__(self, image, **kwargs):
        return self.forward(image)

class RemoveAlpha():

    def __init__(self):
        # some png images have fourth alpha channel with value of 255 everywhere (i.e. opaque). averaging over this adds incorrect offset
        pass

    def forward(self, img):
        if img.shape[2] == 4:
            return img[:, :, :3]
        return img

    def __call__(self, image, **kwargs):
        return self.forward(image)



# def interpolation_lookup(interpolation_str='bilinear'):
#         # PIL.Image.bilinear
#         return getattr(PIL.Image, interpolation_str.upper())



import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils


def demo_rings(root, train, download):
    logging.info('Setting up demo_rings dataset')
    resources = [
        ('https://dl.dropboxusercontent.com/s/4cfb9xbnlkvbkv1/demo_rings_train_catalog.parquet', '4cf7e3eaab46032c374d89c349295aa2'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/xghqujm41ujjlko/demo_rings_test_catalog.parquet', '7b28fc521836e76f794db50c3df4a2e1'),  # test catalog
        ('https://dl.dropboxusercontent.com/s/80s5j6dygtw0gu6/demo_rings_images.tar.gz', '5b3c6c62618bbd3165988c60e36b8365')  # the images
    ]
    images_to_spotcheck = [
        'images/313982/313982_3559.jpg',
        'images/374032/374032_255.jpg'
    ]

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    if train:
        train_catalog_loc = os.path.join(root, 'demo_rings_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc)
    else:
        test_catalog_loc = os.path.join(root, 'demo_rings_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']), axis=1)

    label_cols = ['ring']  # 1 or 0

    logging.info('demo_rings dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    
    df, _ = demo_rings(root='/Users/user/repos/galaxy-datasets/roots/demo_rings', train=True, download=True)
    assert os.path.isfile(df['file_loc'][0])



# import logging
# import os

# import pandas as pd

# from galaxy_datasets.shared import download_utils
# from galaxy_datasets.check_internal_urls import INTERNAL_URLS_EXIST
# if not INTERNAL_URLS_EXIST:
#     raise FileNotFoundError
# from galaxy_datasets.shared import internal_urls

# # TODO could eventually refactor this out of Zoobot as well
# from zoobot.shared import label_metadata

# """
# Downloads DESI galaxies below z<0.1
# Includes GZ DESI labels where available (identical split to )

# Intended for active learning experiments simulated using GZ DESI
# """

# def desi_low_z(root=None, split='train', download=False, train=None):

#     if train is not None:
#         raise ValueError("This dataset has unlabelled data: use split='train', 'test', 'unlabelled' or 'train+unlabelled' rather than train=False etc")

#     assert split in ['train', 'test', 'labelled', 'unlabelled', 'train+unlabelled', 'all']

#     # if root is not None:
#         # 'Legacy Survey cannot be downloaded - ignoring root {}'.format(root)
#         # TODO update for non-manchester users with a manual copy?


    



#     # resources = (
#     #     (internal_urls.legs_train_catalog, '2364a714a1339f020587d374c1838418'),
#     #     (internal_urls.legs_test_catalog, '647bd46f53a06f13eb4df25311ccb9d3'),
#     #     (internal_urls.legs_unlabelled_catalog, '37c9f07b5c058f55d515d2df08ff132a'),
#     #     # and the images, split into 8gb chunks
#     #     (internal_urls.legs_chunk_00, 'd6ca1051b3dd48cfc5c7f0535b403b2d'),
#     #     (internal_urls.legs_chunk_01, 'f258ab647cee076ca66288e25f4a778d'),
#     #     (internal_urls.legs_chunk_02, '7340e212d5eb38d54ea4d89fff93be81'),
#     #     (internal_urls.legs_chunk_03, '33578326983830e2ea5a694757203ae8'),
#     #     (internal_urls.legs_chunk_04, '577150bd970ef802a20cd3cce15f656a'),
#     #     (internal_urls.legs_chunk_05, '1d88458bf9987c7bf5f21301707c9dd8'),
#     #     (internal_urls.legs_chunk_06, '24cf944542e40335f86d7e43468723c0'),
#     #     (internal_urls.legs_chunk_07, '583d92b917bd70670d7860e3836cb4a4')
#     # )
#     resources = (
#         (internal_urls.legs_train_catalog, '2364a714a1339f020587d374c1838418'),
#         (internal_urls.legs_test_catalog, None),
#         (internal_urls.legs_unlabelled_catalog, None),
#         # and the images, split into 8gb chunks
#         (internal_urls.legs_chunk_00, None),
#         (internal_urls.legs_chunk_01, None),
#         (internal_urls.legs_chunk_02, None),
#         (internal_urls.legs_chunk_03, None),
#         (internal_urls.legs_chunk_04, None),
#         (internal_urls.legs_chunk_05, None),
#         (internal_urls.legs_chunk_06, None),
#         (internal_urls.legs_chunk_07, '583d92b917bd70670d7860e3836cb4a4')
#     )

#     if os.path.isdir('/share/nas2'):
#         # hardcoded_catalog_root = '/share/nas2/walml/repos/gz-decals-classifiers/data/decals/presplit_catalogs'
#         pass
#     else:
#         hardcoded_catalog_root = '/home/walml/repos/pytorch-galaxy-datasets/roots/legs'  # catalogs only
#         root = hardcoded_catalog_root
#     downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck=['338503/338503_3840.jpg'], image_dirname='')
#     if download is True:
#         logging.warning('Only downloading catalogs - images are too large to download')
#         downloader.download()

#     label_cols = label_metadata.decals_all_campaigns_ortho_label_cols
#     # useful_columns = label_cols + ['id_str', 'dr8_id', 'brickid', 'objid', 'redshift']

#     train_catalog_loc = os.path.join(root, 'legs_all_campaigns_ortho_dr8_only_train_catalog.parquet')
#     test_catalog_loc = os.path.join(root, 'legs_all_campaigns_ortho_dr8_only_test_catalog.parquet')
#     unlabelled_catalog_loc = os.path.join(root, 'legs_all_campaigns_ortho_dr8_only_unlabelled_catalog.parquet')  # values in label_cols are all 0

#     catalogs = []

#     if 'all' in split or 'train' in split or ('labelled' in split and 'un' not in split):
#         catalogs += [pd.read_parquet(train_catalog_loc)]

#     if 'all' in split or 'test' in split or ('labelled' in split and 'un' not in split):
#         catalogs += [pd.read_parquet(test_catalog_loc)]

#     if 'all' in split or 'unlabelled' in split:
#         catalogs += [pd.read_parquet(unlabelled_catalog_loc)]

#     logging.info('{} catalogs loaded'.format(split))

#     catalog = pd.concat(catalogs, axis=0)
#     catalog = catalog.sample(len(catalog), random_state=42).reset_index(drop=True)

#     catalog['subfolder'] = catalog['brickid'].astype(str)
#     catalog['filename'] = catalog['dr8_id'].astype(str) + '.jpg'
#     catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['subfolder'], x['filename']), axis=1)
#     logging.info(catalog['file_loc'].iloc[0])


#     return catalog, label_cols




import os
import logging

import pandas as pd

from galaxy_datasets.shared import download_utils, label_metadata


def gz_candels(root, train, download):
    logging.info('Setting up gz_candels dataset')
    resources = [
        ('https://dl.dropboxusercontent.com/s/cnjvdinnhh1r1md/candels_ortho_train_catalog.parquet', '90593d1bab79a608cf0e645d6fd8e741'),  # train catalog
        ('https://dl.dropboxusercontent.com/s/y83v1gktw72hs0f/candels_ortho_test_catalog.parquet', '1062993dd8df09684b335678ab3fa8e3'),  # test catalog
        ('https://dl.dropboxusercontent.com/s/d67we9xsn8vyr5k/candels_images.tar.gz', 'b621ee4e650cf084a1a0c1fe5c9d0a21')  # the images
    ]
    images_to_spotcheck = ['COS_9933.jpg', 'UDS_21986.jpg', 'GDS_9405.jpg']

    downloader = download_utils.DatasetDownloader(root, resources, images_to_spotcheck)
    if download is True:
        downloader.download()

    # label_cols = [question + answer for question, answers in candels_ortho_pairs.items() for answer in answers]  # defined below, globally in this script (for imports elsewhere)
    label_cols = label_metadata.candels_ortho_label_cols  # see below

    useful_columns = label_cols + ['filename', 'ra', 'dec']
    if train:
        train_catalog_loc = os.path.join(root, 'candels_ortho_train_catalog.parquet')
        catalog = pd.read_parquet(train_catalog_loc, columns=useful_columns)
    else:
        test_catalog_loc = os.path.join(root, 'candels_ortho_test_catalog.parquet')
        catalog = pd.read_parquet(test_catalog_loc, columns=useful_columns)

    catalog['file_loc'] = catalog.apply(lambda x: os.path.join(downloader.image_dir, x['filename']), axis=1)  # no subfolder

    logging.info('gz_candels dataset ready')
    return catalog, label_cols


if __name__ == '__main__':

    gz_candels(root='roots/gz_candels', train=True, download=True)


import logging
from typing import Union, Callable
import os
import cv2
import json
from itertools import islice


import albumentations as A

import tqdm
import numpy as np
import pandas as pd
from PIL import Image  # necessary to avoid PIL.Image error assumption in web_datasets

from galaxy_datasets.transforms import default_transforms

import webdataset as wds


def catalogs_to_webdataset(dataset_name, wds_dir, label_cols, train_catalog, test_catalog, sparse_label_df=None, divisor=512, overwrite=False):
    for (catalog_name, catalog) in [('train', train_catalog), ('test', test_catalog)]:
        n_shards = len(catalog) // divisor
        logging.info(n_shards)

        catalog = catalog[:n_shards*divisor]
        logging.info(len(catalog))

        save_loc = f"{wds_dir}/{dataset_name}/{dataset_name}_{catalog_name}.tar"  # .tar replace automatically
        
        df_to_wds(catalog, label_cols, save_loc, n_shards=n_shards, sparse_label_df=sparse_label_df, overwrite=overwrite)

    


def df_to_wds(df: pd.DataFrame, label_cols, save_loc: str, n_shards: int, sparse_label_df=None, overwrite=False):

    assert '.tar' in save_loc
    df['id_str'] = df['id_str'].astype(str).str.replace('.', '_')
    if sparse_label_df is not None:
        logging.info(f'Using sparse label df: {len(sparse_label_df)}')
    shard_dfs = np.array_split(df, n_shards)
    logging.info(f'shards: {len(shard_dfs)}. Shard size: {len(shard_dfs[0])}')

    transforms_to_apply = [
        # below, for 224px fast training fast augs setup
        # A.Resize(
        #     height=350,  # now more aggressive, 65% crop effectively
        #     width=350,  # now more aggressive, 65% crop effectively
        #     interpolation=cv2.INTER_AREA  # slow and good interpolation
        # ),
        # A.CenterCrop(
        #     height=224,
        #     width=224,
        #     always_apply=True
        # ),
        # below, for standard training default augs
        # small boundary trim and then resize expecting further 224px crop
        # we want 0.7-0.8 effective crop
        # in augs that could be 0.x-1.0, and here a pre-crop to 0.8 i.e. 340px
        # but this would change the centering
        # let's stick to small boundary crop and 0.75-0.85 in augs

        # turn these off for current euclidized images, already 300x300
        # A.CenterCrop(
        #     height=400,
        #     width=400,
        #     always_apply=True
        # ),
        # A.Resize(
        #     height=300,
        #     width=300,
        #     interpolation=cv2.INTER_AREA  # slow and good interpolation
        # )

        # with GZ Euclid, apply no transforms, the images are saved at native size already
    ]
    transform = A.Compose(transforms_to_apply)
    # transform = None

    for shard_n, shard_df in tqdm.tqdm(enumerate(shard_dfs), total=len(shard_dfs)):
        shard_save_loc = save_loc.replace('.tar', f'_{shard_n}_{len(shard_df)}.tar')
        if overwrite or not(os.path.isfile(shard_save_loc)):
            if sparse_label_df is not None:
                shard_df = pd.merge(shard_df, sparse_label_df, how='left', validate='one_to_one', suffixes=('', '_badlabelmerge'))  # type: ignore # auto-merge

            assert not any(shard_df[label_cols].isna().max()) # type: ignore

            # logging.info(shard_save_loc)
            sink = wds.TarWriter(shard_save_loc)
            for _, galaxy in shard_df.iterrows(): # type: ignore
                try:
                    sink.write(galaxy_to_wds(galaxy, label_cols, transform=transform))
                except Exception as e:
                    logging.critical(galaxy)
                    raise(e)
            sink.close()


def galaxy_to_wds(galaxy: pd.Series, label_cols: Union[list[str],None]=None, metadata_cols: Union[list, None]=None, transform: Union[Callable, None]=None):

    assert os.path.isfile(galaxy['file_loc']), galaxy['file_loc']
    im = cv2.imread(galaxy['file_loc'])
    # cv2 loads BGR for 'history', fix
    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB) 
    assert not np.any(np.isnan(np.array(im))), galaxy['file_loc']

    id_str = str(galaxy['id_str'])

    if transform is not None:
        im = transform(image=im)['image']

    dict_to_save = {
        "__key__": id_str,  # silly wds bug where if __key__ ends .jpg, all keys get jpg. prepended?! use id_str instead
        "image.jpg": im,
    }
    if label_cols is not None:
        dict_to_save['labels.json'] = json.dumps(galaxy[label_cols].to_dict())
    if metadata_cols is not None:
        dict_to_save['metadata.json'] = json.dumps(galaxy[metadata_cols].to_dict())
    return dict_to_save
    
    # huggingface doesn't like empty keys

    # if label_cols is None:
    #     labels = json.dumps({})
    # else:
    #     labels = json.dumps(galaxy[label_cols].to_dict())
  
    # if metadata_cols is None:
    #     metadata = json.dumps({})
    # else:
    #     
    
    # return {
    #     "image.jpg": im,
    #     "labels.json": labels,
    #     "metadata.json": metadata
    # }


# just for debugging
def load_wds_directly(wds_loc, max_to_load=3):

    dataset = wds.WebDataset(wds_loc) \
    .decode("rgb")

    if max_to_load is not None:
        sample_iterator = islice(dataset, 0, max_to_load)
    else:
        sample_iterator = dataset
    for sample in sample_iterator:
        logging.info(sample['__key__'])     
        logging.info(sample['image.jpg'].shape)  # .decode(jpg) converts to decoded to 0-1 RGB float, was 0-255
        logging.info(type(sample['labels.json']))  # automatically decoded


# just for debugging
def load_wds_with_augmentation(wds_loc):

    augmentation_transform = default_transforms()  # A.Compose object
    def do_transform(img):
        return np.transpose(augmentation_transform(image=np.array(img))["image"], axes=[2, 0, 1]).astype(np.float32)

    dataset = wds.WebDataset(wds_loc) \
        .decode("rgb") \
        .to_tuple('image.jpg', 'labels.json') \
        .map_tuple(do_transform, identity)
    
    for sample in islice(dataset, 0, 3):
        logging.info(sample[0].shape)
        logging.info(sample[1])


def identity(x):
    # no lambda to be pickleable
    return x



def make_mock_wds(save_dir: str, label_cols: list, n_shards: int, shard_size: int):
    counter = 0
    shards = [os.path.join(save_dir, f'mock_shard_{shard_n}_{shard_size}.tar') for shard_n in range(n_shards)]
    for shard in shards:
        sink = wds.TarWriter(shard)
        for galaxy_n in range(shard_size):
            data = {
                "__key__": f'id_{galaxy_n}',
                "image.jpg": (np.random.rand(424, 424)*255.).astype(np.uint8),
                "labels.json": json.dumps(dict(zip(label_cols, [np.random.randint(low=0, high=10) for _ in range(len(label_cols))])))
            }
            sink.write(data)
            counter += 1
    print(counter)
    return shards


if __name__ == '__main__':

    save_dir = '/home/walml/repos/temp'
    from galaxy_datasets.shared import label_metadata
    label_cols = label_metadata.decals_all_campaigns_ortho_label_cols

    make_mock_wds(save_dir, label_cols, n_shards=4, shard_size=512)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="galaxy-datasets",
    version="0.0.23",
    author="Mike Walmsley",
    author_email="walmsleymk1@gmail.com",
    description="Galaxy Zoo datasets for PyTorch/TensorFlow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwalmsley/galaxy-datasets",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: GPU :: NVIDIA CUDA"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.9",  # zoobot has 3.9,
    install_requires=[
        'numpy',
        'pandas',
        'pyarrow',
        'requests',
        'omegaconf',  # for config
        'datasets'  # for HF typing
    ],
    extras_require={
        # these are lower than zoobot's reqs
        'pytorch': [
            'torch >= 1.10.1',
            'torchvision >= 0.11.2',
            'torchaudio >= 0.10.1',
            'pytorch-lightning >=2.0',
            'albumentations <= 2.0'  # torchvision now preferred
        ],
        'tensorflow': [ # will be deprecated
            'tensorflow >= 2.10.0',
            'protobuf <= 3.19'  # tensorflow incompatible above this (usually resolved by pip automatically)
        ]
        # if you need both, specify both
    }
)



import logging
from functools import partial
from typing import Optional, Union

import numpy as np
import pandas as pd
import pytorch_lightning as pl
from datasets import load_dataset
from sklearn.model_selection import train_test_split

# import torch
from torch.utils.data import DataLoader
import datasets as hf_datasets

from galaxy_datasets.pytorch import galaxy_dataset
from galaxy_datasets.transforms import default_transforms


# https://pytorch-lightning.readthedocs.io/en/stable/extensions/datamodules.html
class GalaxyDataModule(pl.LightningDataModule):
    # takes generic catalogs (which are already downloaded and happy),
    # splits if needed, and creates generic datasets->dataloaders etc
    # easy to make dataset-specific default transforms if desired
    def __init__(
        self,
        label_cols,
        # provide full catalog for automatic split, or...
        catalog=None,
        train_fraction=0.7,
        val_fraction=0.1,
        test_fraction=0.2,
        # provide train/val/test catalogs for your own previous split
        train_catalog: Optional[pd.DataFrame] = None,
        val_catalog: Optional[pd.DataFrame] = None,
        test_catalog: Optional[pd.DataFrame] = None,
        predict_catalog: Optional[pd.DataFrame] = None,
        # augmentation params (sensible supervised defaults)
        greyscale=False,
        # album=False,  # now True always
        crop_scale_bounds=(0.7, 0.8),
        crop_ratio_bounds=(0.9, 1.1),
        resize_after_crop=224,
        custom_albumentation_transform=None,  # will override the settings above. If tuple, assume (train, test) transforms
        custom_torchvision_transform=None,  # similarly
        # the above will soon be replaced with
        # transform_cfg=None,
        # train_transform_cfg=None,
        # inference_transform_cfg=None,
        # hardware params
        batch_size=256,  # careful - will affect final performance
        use_memory=False,  # deprecated
        num_workers=4,
        prefetch_factor=4,
        seed=42,
    ):
        super().__init__()

        if catalog is not None:  # catalog provided, should not also provide explicit split catalogs
            assert train_catalog is None
            assert val_catalog is None
            assert test_catalog is None
        else:  # catalog not provided, must provide explicit split catalogs - at least one
            assert (
                (train_catalog is not None)
                or (val_catalog is not None)
                or (test_catalog is not None)
                or (predict_catalog is not None)
            )
            # see setup() for how having only some explicit catalogs is handled

        self.label_cols = label_cols

        self.catalog = catalog
        self.train_catalog = train_catalog
        self.val_catalog = val_catalog
        self.test_catalog = test_catalog
        self.predict_catalog = predict_catalog

        self.batch_size = batch_size

        self.use_memory = use_memory
        if self.use_memory:
            raise NotImplementedError

        self.num_workers = num_workers
        self.seed = seed

        assert np.isclose(train_fraction + val_fraction + test_fraction, 1.0)
        self.train_fraction = train_fraction
        self.val_fraction = val_fraction
        self.test_fraction = test_fraction

        if self.num_workers == 0:
            logging.warning(
                "num_workers=0, setting prefetch=None and timeout=0 as no multiprocessing"
            )
            self.prefetch_factor = None
            self.dataloader_timeout = 0
        else:
            self.prefetch_factor = prefetch_factor
            self.dataloader_timeout = 600  # seconds aka 10 mins

        logging.info("Num workers: {}".format(self.num_workers))
        logging.info("Prefetch factor: {}".format(self.prefetch_factor))

        self.custom_albumentation_transform = custom_albumentation_transform
        self.custom_torchvision_transform = custom_torchvision_transform
        self.resize_after_crop = resize_after_crop
        self.crop_scale_bounds = crop_scale_bounds
        self.crop_ratio_bounds = crop_ratio_bounds
        self.greyscale = greyscale

        if custom_torchvision_transform is not None:
            logging.info("Using custom torchvision transform for augmentations")
            self.transform_with_torchvision()
        else:
            logging.info("Using albumentations transform for augmentations")
            self.transform_with_albumentations()

    def transform_with_torchvision(self):
        # TODO
        # galaxy_datasets loads PIL images, likely need PILToTensor() in any custom torchvision transform. review.
        if isinstance(self.custom_torchvision_transform, tuple):
            logging.info("Using different torchvision transforms for train and test")
            assert len(self.custom_torchvision_transform) == 2
            self.train_transform = self.custom_torchvision_transform[0]
            self.test_transform = self.custom_torchvision_transform[1]
        else:
            self.train_transform = self.custom_torchvision_transform
            self.test_transform = self.custom_torchvision_transform

    def transform_with_albumentations(self):
        import albumentations as A

        if self.custom_albumentation_transform is not None:
            logging.warning('Custom albumentation transforms will be deprecated in future versions. Use torchvision transforms instead.')
            if isinstance(self.custom_albumentation_transform, tuple):
                logging.info("Using different albumentations transforms for train and test")
                assert len(self.custom_albumentation_transform) == 2
                assert isinstance(self.custom_albumentation_transform[0], A.Compose)
                self.train_transform = partial(
                    do_transform,
                    transforms_to_apply=self.custom_albumentation_transform[0],
                )
                self.test_transform = partial(
                    do_transform,
                    transforms_to_apply=self.custom_albumentation_transform[1],
                )
            else:
                logging.info("Using the same custom albumentations transforms for train and test")
                self.train_transform = partial(
                    do_transform,
                    transforms_to_apply=self.custom_albumentation_transform,
                )
                self.test_transform = partial(
                    do_transform,
                    transforms_to_apply=self.custom_albumentation_transform,
                )
        else:
            logging.info("Using basic albumentations transforms for augmentations")
            # gives a transforms = Compose() object
            transforms_to_apply = default_transforms(
                crop_scale_bounds=self.crop_scale_bounds,
                crop_ratio_bounds=self.crop_ratio_bounds,
                resize_after_crop=self.resize_after_crop,
                pytorch_greyscale=self.greyscale,
            )
            self.train_transform = partial(do_transform, transforms_to_apply=transforms_to_apply)
            self.test_transform = partial(do_transform, transforms_to_apply=transforms_to_apply)

    # only called on main process
    def prepare_data(self):
        pass  # could include some basic checks

    # called on every gpu

    def setup(self, stage: Optional[str] = None):

        self.specify_catalogs(stage)

        # Assign train/val datasets for use in dataloaders
        # assumes dataset_class has these standard args
        if stage == "fit" or stage is None:
            self.train_dataset = galaxy_dataset.GalaxyDataset(
                catalog=self.train_catalog,
                label_cols=self.label_cols,
                transform=self.train_transform,
            )
            self.val_dataset = galaxy_dataset.GalaxyDataset(
                catalog=self.val_catalog,
                label_cols=self.label_cols,
                transform=self.test_transform,
            )

        # Assign test dataset for use in dataloader(s)
        if stage == "test" or stage is None:
            self.test_dataset = galaxy_dataset.GalaxyDataset(
                catalog=self.test_catalog,
                label_cols=self.label_cols,
                transform=self.test_transform,
            )

        if (
            stage == "predict"
        ):  # not set up by default with stage=None, only if explicitly requested
            if self.predict_catalog is None:
                raise ValueError(
                    "Attempting to predict, but GalaxyDataModule was init without a predict_catalog arg. init with GalaxyDataModule(predict_catalog=some_catalog, ...)"
                )
            self.predict_dataset = galaxy_dataset.GalaxyDataset(
                catalog=self.predict_catalog,
                label_cols=self.label_cols,
                transform=self.test_transform,
            )

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
            prefetch_factor=self.prefetch_factor,
            timeout=self.dataloader_timeout,
            drop_last=True,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
            prefetch_factor=self.prefetch_factor,
            timeout=self.dataloader_timeout,
        )

    def test_dataloader(self):
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
            prefetch_factor=self.prefetch_factor,
            timeout=self.dataloader_timeout,
        )

    def predict_dataloader(self):
        return DataLoader(
            self.predict_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=True,
            persistent_workers=self.num_workers > 0,
            prefetch_factor=self.prefetch_factor,
            timeout=self.dataloader_timeout,
        )

    def specify_catalogs(self, stage):
        if self.catalog is not None:
            # will split the catalog into train, val, test here
            self.train_catalog, hidden_catalog = train_test_split(
                self.catalog, train_size=self.train_fraction, random_state=self.seed
            )
            self.val_catalog, self.test_catalog = train_test_split(
                hidden_catalog,
                train_size=self.val_fraction / (self.val_fraction + self.test_fraction),
                random_state=self.seed,
            )
            del hidden_catalog
        else:
            # assume you have passed pre-split catalogs
            # (maybe not all, e.g. only a test catalog, or only train/val catalogs)
            if stage == "predict":
                assert self.predict_catalog is not None
            elif stage == "test":
                # only need test
                assert self.test_catalog is not None
            elif stage == "fit":
                # only need train and val
                assert self.train_catalog is not None
                assert self.val_catalog is not None
            else:
                # need all three (predict is still optional)
                assert self.train_catalog is not None
                assert self.val_catalog is not None
                assert self.test_catalog is not None
            # (could write this shorter but this is clearest)


def do_transform(img, transforms_to_apply):
    # albumentations expects np array, and returns dict keyed by "image"
    # transforms_to_apply should return tensor in BCHW (torch style). 
    # true by default since default_transforms now uses to_tensor=True
    # Albumentations doesn't do this by default, so add transpose/to_tensor if using custom albumentations
    return transforms_to_apply(image=np.array(img))["image"]


# https://pytorch-lightning.readthedocs.io/en/stable/extensions/datamodules.html
class HF_GalaxyDataModule(GalaxyDataModule):
    def __init__(
        self,
        dataset: hf_datasets.Dataset,
        label_cols: list,
        train_transform,
        test_transform,
        target_transform=None,
        # hardware params
        batch_size=256,  # careful - will affect final performance
        num_workers=4,
        prefetch_factor=4,
        seed=42,
    ):
        # super().__init__()
        pl.LightningDataModule.__init__(self)

        self.batch_size = batch_size

        self.num_workers = num_workers
        self.seed = seed
        self.label_cols = label_cols
        self.dataset = dataset
        self.train_transform = train_transform
        self.test_transform = test_transform
        self.target_transform = target_transform

        if self.num_workers == 0:
            logging.warning(
                "num_workers=0, setting prefetch=None and timeout=0 as no multiprocessing"
            )
            self.prefetch_factor = None
            self.dataloader_timeout = 0
        else:
            self.prefetch_factor = prefetch_factor
            self.dataloader_timeout = 600  # seconds aka 10 mins

        logging.info("Num workers: {}".format(self.num_workers))
        logging.info("Prefetch factor: {}".format(self.prefetch_factor))

        # self.prepare_data_per_node = False  # run prepare_data below only on master node (and one process)

    # only called on main process
    def prepare_data(self):
        pass

    # called on every gpu
    def setup(self, stage: Optional[str] = None):
        # Assign train/val datasets for use in dataloaders
        # assumes dataset_class has these standard args
        if stage == "fit" or stage is None:
            if "val" in self.dataset.keys():
                # life is good, we have all the splits we need
                train_dataset_hf = self.dataset['train']
                val_dataset_hf = self.dataset['val']
            else:
                # life is less good, we need to make a val split from train
                # (not from test, to avoid overfitting via early stopping)
                logging.warning('No premade validation split, creating from 20%% of train dataset')
                train_datasetdict = self.dataset["train"].train_test_split(test_size=0.2, shuffle=True)
                train_dataset_hf = train_datasetdict["train"]
                val_dataset_hf = train_datasetdict['test']  # actually used as val
                del train_datasetdict

            self.train_dataset = galaxy_dataset.HF_GalaxyDataset(
                    dataset=train_dataset_hf,
                    label_cols=self.label_cols,
                    transform=self.train_transform,
                    target_transform=self.target_transform,
            )
            self.val_dataset = galaxy_dataset.HF_GalaxyDataset(
                dataset=val_dataset_hf,
                label_cols=self.label_cols,
                transform=self.test_transform,
                target_transform=self.target_transform,
            )

        # Assign test dataset for use in dataloader(s)
        if stage == "test" or stage is None:
            self.test_dataset = galaxy_dataset.HF_GalaxyDataset(
                dataset=self.dataset['test'],
                label_cols=self.label_cols,
                transform=self.test_transform,
                target_transform=self.target_transform,
            )




from galaxy_datasets.shared import gz2, gz_candels, gz_decals_5, gz_hubble, demo_rings, galaxy_mnist, tidal

from galaxy_datasets.pytorch import galaxy_dataset


# TODO could refactor these into same class if needed

class GZCandels(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = gz_candels(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)


class GZDecals5(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = gz_decals_5(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)



class GZ2(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = gz2(root, train, download)  # no train arg

        super().__init__(catalog, label_cols, transform, target_transform)


class GZHubble(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = gz_hubble(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)


class DemoRings(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = demo_rings(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)

class GalaxyMNIST(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

        catalog, label_cols = galaxy_mnist(root, train, download)

        super().__init__(catalog, label_cols, transform, target_transform)

class Tidal(galaxy_dataset.GalaxyDataset):
    
    def __init__(self, root, train=True, download=False, transform=None, target_transform=None, label_mode='coarse'):

        catalog, label_cols = tidal(root, train, download, label_mode=label_mode)

        super().__init__(catalog, label_cols, transform, target_transform)



from galaxy_datasets import check_internal_urls
if check_internal_urls.INTERNAL_URLS_EXIST:
    from galaxy_datasets.shared import gz_desi, gz_rings, gz_h2o, gz_ukidss, gz_jwst

    class GZDesi(galaxy_dataset.GalaxyDataset):
        
        def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

            catalog, label_cols = gz_desi(root, train, download)

            super().__init__(catalog, label_cols, transform, target_transform)

    class GZRings(galaxy_dataset.GalaxyDataset):
        
        def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

            catalog, label_cols = gz_rings(root, train, download)

            super().__init__(catalog, label_cols, transform, target_transform)

    class GZH2O(galaxy_dataset.GalaxyDataset):
        
        def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

            catalog, label_cols = gz_h2o(root, train, download)

            super().__init__(catalog, label_cols, transform, target_transform)


    class GZUKIDSS(galaxy_dataset.GalaxyDataset):
        
        def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

            catalog, label_cols = gz_ukidss(root, train, download)

            super().__init__(catalog, label_cols, transform, target_transform)


    class GZJWST(galaxy_dataset.GalaxyDataset):
        
        def __init__(self, root, train=True, download=False, transform=None, target_transform=None):

            catalog, label_cols = gz_jwst(root, train, download)

            super().__init__(catalog, label_cols, transform, target_transform)



# temporarily deprecated
# class Legs(galaxy_dataset.GalaxyDataset):
    
#     # based on https://pytorch.org/vision/stable/generated/torchvision.datasets.STL10.html
#     def __init__(self, root=None, split='train', download=False, transform=None, target_transform=None, train=None):
#         # train=None is just an exception-raising parameter to avoid confused users using the train=False api

#         catalog, label_cols = legs(root, split, download, train)

#         # paths are not adjusted as cannot be downloaded
#         # catalog = _temp_adjust_catalog_paths(catalog)
#         # catalog = adjust_catalog_dtypes(catalog, label_cols)

#         super().__init__(catalog, label_cols, transform, target_transform)
